#-----------------------------------
# Imports
#-----------------------------------

from typing import Any

from PyQt5.QtCore import QCoreApplication, Qt, pyqtSlot
from PyQt5.QtWidgets import QLabel, QFormLayout, QComboBox, QWidget, QVBoxLayout, QHBoxLayout, QPushButton, QGridLayout

from muphyn.packages.core.plci_core_scheduler_params import SchedulerParams
from muphyn.packages.core.scheduler_library.plci_core_schedulers_libraries import SchedulersLibraries
from muphyn.packages.core.scheduler_library.scheduler_library_data import SchedulerData
from muphyn.packages.interface.models.dbl_spin_box import DblSpinBox
from muphyn.packages.interface.dialogs.abstract_dialog import AbstractDialog
from muphyn.packages.interface.models.editable_models.scheduler_model import SchedulerModel
from muphyn.packages.interface.models.editable_models.simulation_model import SimulationModel
from muphyn.packages.interface.widgets.BaseWidgets.Button import PlainButton
from muphyn.packages.interface.widgets.BaseWidgets.Label import PropertyLabel
from muphyn.packages.interface.utils.interface_constants import ApplicationWindowTitle

#-----------------------------------
# Class
#-----------------------------------

class SimulationParametersDialog (AbstractDialog) :
    """Est la classe de dialogue qui permet de modifier les paramètres de la simulation."""
    
    # -------------
    # Constructors
    # -------------

    def __init__ (self, dialog_holder : Any, simulation_model : SimulationModel, schedulers_libraries : SchedulersLibraries) :

        AbstractDialog.__init__(self, dialog_holder, 'simulation_parameters_dialog', QCoreApplication.translate(u'dlg_simulation_parameter', f'{ApplicationWindowTitle} - Simulation parameters', None))

        self.setFixedSize(370, 160)
        self.init_ui()
        self.tranlsate_ui()

        self._simulation_model = None
        self._last_selected_index = -1
        self._schedulers_libraries = schedulers_libraries

        for scheduler_data in self._schedulers_libraries.schedulers :
            self._cmb_scheduler.addItem(scheduler_data.__str__(), scheduler_data)

        self.simulation_model = simulation_model

    # -------------
    # Properties
    # -------------

    @property 
    def simulation_model (self) -> SimulationModel :
        """Permet de récuperer le modèle de simulation."""
        return self._simulation_model

    @simulation_model.setter
    def simulation_model (self, simulation_model_ : SimulationModel) -> None :
        """Permet de modifier le modèle de simulation."""

        self._simulation_model = simulation_model_

        if self._simulation_model is None : 
            self._btn_accept.setEnabled(False)
            
            self._spn_simulation_tick.setEnabled(False)
            self._spn_simulation_tick.setValue(0)

            self._spn_simulation_time.setEnabled(False)
            self._spn_simulation_time.setValue(0)

            self._cmb_scheduler.setEnabled(False)
            if self._cmb_scheduler.count() > 0 :
                self._cmb_scheduler.setCurrentIndex(0)

        else : 
            
            self._spn_simulation_tick.setEnabled(True)
            self._spn_simulation_time.setEnabled(True)
            self._cmb_scheduler.setEnabled(True)
            
            if self._simulation_model.scheduler_model is None :
                self._spn_simulation_tick.setValue(10)
                self._spn_simulation_time.setValue(0.01)

            else : 
                if self._simulation_model.scheduler_model.params is None : 
                    self._spn_simulation_tick.setValue(10)
                    self._spn_simulation_time.setValue(0.01)
                    
                else :
                    self._spn_simulation_tick.setValue(self._simulation_model.scheduler_model.params.step_time)
                    self._spn_simulation_time.setValue(self._simulation_model.scheduler_model.params.stop_time)


            if self._cmb_scheduler.count() > 0 :

                self._cmb_scheduler.setCurrentIndex(0)

                if not(self._simulation_model.scheduler_model is None) :

                    for index in range(self._cmb_scheduler.count()) :
                        item_text = self._cmb_scheduler.itemText(index)
                        if  item_text == self._simulation_model.scheduler_model.__str__() :
                            self._last_selected_index = index
                            self._cmb_scheduler.setCurrentIndex(index)
                            break
        
        self.check_changed()

    # -------------
    # Methods
    # -------------

    @pyqtSlot()
    def button_accept_clicked (self) -> None :
        """Est la méthode appelée quand l'utilisateur clique sur le bouton accepter du dialogue."""
        
        scheduler_data : SchedulerData = self._cmb_scheduler.currentData() 
        scheduler_params = SchedulerParams(self._spn_simulation_time.value(), self._spn_simulation_tick.value())
        self._value = SchedulerModel(scheduler_data.scheduler_library, scheduler_data.scheduler_name, scheduler_params)
        self.close()

    @pyqtSlot()
    def button_cancel_clicked (self) -> None : 
        """Est la méthode appelée quand l'utilisateur clique sur le bouton annuler du dialogue."""
        self.close()

    @pyqtSlot()
    def check_changed (self) -> None : 
        """Est la méthode appelée quand l'utilisateur modifie un élément graphique. La méthode débloque alors le bouton accepter si il y a quelque chose de modifié."""

        if self.simulation_model is None or not(self._cmb_scheduler.count() > 0) :
            self._btn_accept.setEnabled(False)
            return

        if self.simulation_model.scheduler_model is None : 

            if not(self._cmb_scheduler.currentIndex() == -1) :
                self._btn_accept.setEnabled(True)
            
            else :
                self._btn_accept.setEnabled(False)

            return

        if self.simulation_model.scheduler_model.params is None : 
            self._btn_accept.setEnabled(True)
            return 

        # If Simulation Duration time is zero or Step time is zero → invalid value
        if self._spn_simulation_tick.value() == 0 or self._spn_simulation_time.value() == 0 :
            self._btn_accept.setEnabled(False)
            return

        # If Step time is greater than simulation time → invalid value
        if self._spn_simulation_time.value()  < self._spn_simulation_tick.value() : 
            self._btn_accept.setEnabled(False)
            return

        # If Scheduler is different from set parameters → enable button
        if self.simulation_model.scheduler_model.completeName != self._cmb_scheduler.currentText():
            self._btn_accept.setEnabled(True)
            return

        # If Step Time is different from set parameters → enable button
        if self.simulation_model.scheduler_model.params.step_time != self._spn_simulation_tick.value() :
            self._btn_accept.setEnabled(True)
            return

        # If Stop Time is different from set parameters → enable button
        if self.simulation_model.scheduler_model.params.stop_time != self._spn_simulation_time.value() :
            self._btn_accept.setEnabled(True)
            return

        # Disable accept button
        self._btn_accept.setEnabled(False)

    def init_ui (self) -> None :
        """Est la méthode appelée pour initialiser les éléments de l'interface graphique."""

        if self.objectName() is None : 
            self.setObjectName('dlg_simulation_parameter')

        mainLayout = QVBoxLayout()

        # Options Layout
        optionsLayout = QGridLayout()

        # Buttons Layout
        buttonsLayout = QGridLayout()
        buttonsLayout.setColumnMinimumWidth(1, 90)
        buttonsLayout.setColumnMinimumWidth(2, 90)
        buttonsLayout.setHorizontalSpacing(5)

        # Scheduler
        self._lbl_scheduler : PropertyLabel = PropertyLabel()
        self._cmb_scheduler : QComboBox = QComboBox()
        self._cmb_scheduler.currentIndexChanged.connect(self.check_changed)

        # Time Value
        self._lbl_simulation_time : PropertyLabel = PropertyLabel()
        self._spn_simulation_time : DblSpinBox = DblSpinBox()
        self._spn_simulation_time.setMinimum(0)
        self._spn_simulation_time.valueChanged.connect(self.check_changed)

        # Step Time Value
        self._lbl_simulation_tick : PropertyLabel = PropertyLabel()
        self._spn_simulation_tick : DblSpinBox = DblSpinBox()
        self._spn_simulation_tick.setMinimum(0)
        self._spn_simulation_tick.valueChanged.connect(self.check_changed)

        # Accept button
        self._btn_accept : QPushButton = PlainButton()
        self._btn_accept.clicked.connect(self.button_accept_clicked)

        # Cancel Button
        self._btn_cancel : QPushButton = QPushButton()
        self._btn_cancel.clicked.connect(self.button_cancel_clicked)
        
        # Add option widgets to layout
        optionsLayout.addWidget(self._lbl_scheduler, 0, 0)
        optionsLayout.addWidget(self._cmb_scheduler, 0, 1)
        optionsLayout.addWidget(self._lbl_simulation_time, 1, 0)
        optionsLayout.addWidget(self._spn_simulation_time, 1, 1)
        optionsLayout.addWidget(self._lbl_simulation_tick, 2, 0)
        optionsLayout.addWidget(self._spn_simulation_tick, 2, 1)
        optionsLayout.setColumnStretch(1, 1)

        # Add buttons to layout
        buttonsLayout.addWidget(self._btn_accept, 0, 1)
        buttonsLayout.addWidget(self._btn_cancel, 0, 2)
        buttonsLayout.setColumnStretch(0, 1)

        # Add Options Layout to Main Layout
        mainLayout.addLayout(optionsLayout, 1)

        # Add Buttons Layout to Main Layout
        mainLayout.addLayout(buttonsLayout)

        self.setLayout(mainLayout)

    def tranlsate_ui (self) -> None : 
        """Est la méthode pour traduire les éléments de l'interface graphique."""

        self._lbl_scheduler.setText(QCoreApplication.translate(self.objectName(), u'Scheduler : ', None))
        self._lbl_simulation_time.setText(QCoreApplication.translate(self.objectName(), u'Duration time [s] :', None))
        self._lbl_simulation_tick.setText(QCoreApplication.translate(self.objectName(), u'Step time [s] :', None))
        self._btn_accept.setText(QCoreApplication.translate(self.objectName(), u'Accept', None))
        self._btn_cancel.setText(QCoreApplication.translate(self.objectName(), u'Cancel', None))


import numpy as np
from typing import List

def print_eq_laplace (num_vect : List[float], denum_vect : List[float]) -> None : 
    eq = 'Y * ('
    order_denum = len(denum_vect) - 1
    order_num = len(num_vect) - 1

    for i, current_coeff in enumerate(denum_vect) :
        
        current_order = order_denum - i

        if current_coeff == 0 :
            continue

        if not(i == 0) :
            
            if current_coeff < 0 :
                eq = eq + ' - '
            else :
                eq = eq + ' + '

        else :
            if current_coeff < 0 :
                eq = eq + ' - '

        if current_order > 1 :
            eq = eq + str(np.abs(current_coeff)) + ' s**' + str(current_order)

        elif current_order == 1 :
            eq = eq + str(np.abs(current_coeff)) + ' s'

        elif current_order == 0 : 
            eq = eq + str(np.abs(current_coeff))

    eq = eq + ') / ('

    for i, current_coeff in enumerate(num_vect) :
        
        current_order = order_num - i
        
        if current_coeff == 0 :
            continue
        
        if not(i == 0) :
            
            if current_coeff < 0 :
                eq = eq + ' - '
            else :
                eq = eq + ' + '

        else :
            if current_coeff < 0 :
                eq = eq + ' - '

        if current_order > 1 :
            eq = eq + str(np.abs(current_coeff)) + ' s**' + str(current_order)

        elif current_order == 1 :
            eq = eq + str(np.abs(current_coeff)) + ' s'

        elif current_order == 0 : 
            eq = eq + str(np.abs(current_coeff))

    return eq + ') = U'


#-----------------------------------
# Imports
#-----------------------------------

import os
import time
import pandas as pd
from typing import Any

from muphyn.packages.core.plci_core_scheduler import Scheduler
from muphyn.packages.core.utils.file import CreateDirectory

#-----------------------------------
# Methods
#-----------------------------------

def _execute_box (scheduler: Scheduler, event, box) -> None :
    
    new_events = box.function(event)

    if scheduler.is_scheduler_exception(new_events) :
        return new_events

    for event in new_events :
        
        scheduler.append_event(event)

        if event.signal in scheduler.diagram.linked_signals :
            for signal in scheduler.diagram.linked_signals[event.signal] :
                scheduler.append_event(scheduler.construct_signal_event(signal, event.box, event.new_signal_data))

def _default_scheduler_method (scheduler: Scheduler) -> Any :
    
    b_bis = None
    event = None
    b = None

    # Init pandas table 
    timings = {
        "box": [], "box_start_time": [], "box_end_time": [], "scheduler_start_exception_time": [], "scheduler_end_exception_time": []
    }

    # Global start time
    global_start_time = time.perf_counter()

    try :

        while scheduler.should_simulation_continue :
            
            for b in scheduler.diagram.boxes :
                timings["box"].append(f"{b.name}_{b.index}")
                
                # Execute box
                timings["box_start_time"].append(time.perf_counter())
                new_events = _execute_box(scheduler, scheduler.construct_event(), b)

                # Scheduler Exception time
                scheduler_exception_start_time = time.perf_counter()
                timings["box_end_time"].append(scheduler_exception_start_time)
                timings["scheduler_start_exception_time"].append(scheduler_exception_start_time)

                # Scheduler Exception
                if scheduler.is_scheduler_exception(new_events) :
                    return new_events
                    
                # Scheduler Exception time
                timings["scheduler_end_exception_time"].append(time.perf_counter())

                while scheduler.are_events_left :
                    
                    event = scheduler.get_event(0)
                    event.signal.data = event.new_signal_data

                    if event.signal in scheduler.diagram.box_inputs :
                    
                        for b_bis in scheduler.diagram.box_inputs[event.signal] :
                            timings["box"].append(f"{b_bis.name}_{b_bis.index}")
                            
                            # Execute box
                            timings["box_start_time"].append(time.perf_counter())
                        
                            new_events = _execute_box(scheduler, scheduler.construct_event(event), b_bis)

                            # Scheduler Exception time
                            scheduler_exception_start_time = time.perf_counter()
                            timings["box_end_time"].append(scheduler_exception_start_time)
                            timings["scheduler_start_exception_time"].append(scheduler_exception_start_time)
                            
                            if scheduler.is_scheduler_exception(new_events) :
                                return new_events

                            # Scheduler Exception time
                            timings["scheduler_end_exception_time"].append(time.perf_counter())
                            
                    scheduler.remove_event(event)

            scheduler.stepping_time()

    except Exception as err :
        return scheduler.construct_exception(b, b_bis, event, err)

    
    # Global end time
    global_end_time = time.perf_counter()

    # Convert to dataframe
    pd_timings = pd.DataFrame(timings)

    # 
    pd_timings["box_duration"] = pd_timings["box_end_time"] - pd_timings["box_start_time"]
    pd_timings["scheduler_duration"] = pd_timings["scheduler_end_exception_time"] - pd_timings["scheduler_start_exception_time"]

    # Create directory if doesn't exists
    directory = "LOGS/Time"
    CreateDirectory(directory)

    # Save timing report
    pd_timings.to_csv(os.path.join(directory, f"Log_time_{time.strftime('%Y_%m_%j_%H_%M_%S')}.csv"), sep=";", decimal=",")


#-----------------------------------
# Imports
#-----------------------------------

from typing import Dict

from muphyn.packages.core.scheduler_library.scheduler_library_data import SchedulerData

#-----------------------------------
# Class
#-----------------------------------

class AbstractSchedulerLibraryImporter : 
    """Est la classe commune des classes capable d'importer des planificateur."""

    # -------------
    # Constructors
    # -------------

    def __init__ (self) :
        ...

    # -------------
    # Methods
    # -------------

    def import_scheduler_data (self, path : str, file_name : str, absolute_yaml_file : str, schedulers : Dict[str, SchedulerData]) -> Dict[str, SchedulerData] :
        ...


#-----------------------------------
# Imports
#-----------------------------------

from PyQt5 import QtGui
from PyQt5.QtCore import QCoreApplication, Qt, pyqtSlot
from PyQt5.QtWidgets import QComboBox, QLabel, QLineEdit, QSlider
from muphyn.packages.interface.models.dbl_spin_box import DblSpinBox
from muphyn.packages.interface.models.signals_model.signal_link_model import SignalLinkModel
from muphyn.packages.interface.models.links_model.link_type import LinkType
from muphyn.packages.interface.properties_pages.abstract_properties_editor import AbstractPropertiesEditor
from muphyn.packages.interface.widgets.BaseWidgets.Label import PropertyLabel

#-----------------------------------
# Class
#-----------------------------------

sldr_factor = 100.0

class SignalPropertiesEditor (AbstractPropertiesEditor) : 
    """Est la classe qui affiche une page capable de modifier les propriétés d'un lien."""

    # -------------
    # Constructors
    # -------------

    def __init__ (self, signal_model : SignalLinkModel) :
        
        AbstractPropertiesEditor.__init__(self, signal_model)

        self._value_semaphore : bool = True
        self._link_value_semaphore : bool = False
        self._signal_model : SignalLinkModel = None
        self.signal_model = signal_model
        self._value_semaphore : bool = False
        
    # -------------
    # Properties
    # -------------
    
    @property
    def signal_model (self) -> SignalLinkModel :
        """Permet de récuperer le signal modifié par la fenêtre de propriétés."""
        return self._signal_model
    
    @signal_model.setter
    def signal_model (self, signal_model_ : SignalLinkModel) -> None :
        """Permet de modifier le signal modifié par la fenêtre de propriétés."""

        if not(self._signal_model is None) :
            self._signal_model.param_changed.disconnect(self.signal_model_param_changed)

        self._signal_model = signal_model_
        self._model = signal_model_

        if self._signal_model == None :
            self._fld_texte.setEnabled(False)
            self._fld_texte.setText('')
            self._cmb_color.setEnabled(False)
            self._cmb_color.setCurrentIndex(0)
            self._cmb_link_type.setEnabled(False)
            self._cmb_link_type.setCurrentIndex(0)
            self._spn_link_value.setEnabled(False)
            self._spn_link_value.setValue(0)
            self._sldr_link_value.setEnabled(False)
            self._sldr_link_value.setValue(0)
            self._lbl_data_type_value.setText('')

        else :
            self._fld_texte.setEnabled(True)
            self._fld_texte.setText(self._signal_model.text)
            
            self._cmb_color.setEnabled(False)
            self._cmb_color.setCurrentIndex(0)

            self._cmb_link_type.setEnabled(True)
            if self._signal_model.link_type == LinkType.SQUARE :
                self._cmb_link_type.setCurrentIndex(0)
            elif self._signal_model.link_type == LinkType.CURVED :
                self._cmb_link_type.setCurrentIndex(1)

            self._link_value_semaphore : bool = True
            self._spn_link_value.setEnabled(True)
            self._spn_link_value.setValue(self._signal_model.link_value)
            
            self._sldr_link_value.setEnabled(True)
            self._sldr_link_value.setValue(int(self._signal_model.link_value * sldr_factor))
            self._link_value_semaphore : bool = False

            self._lbl_data_type_value.setText(self._signal_model.data_type.__str__())
            
            self._signal_model.param_changed.connect(self.signal_model_param_changed)

    # -------------
    # Methods
    # -------------

    def signal_model_param_changed (self, signal_model, param_name, old_value, new_value) -> None :
        """Est la méthode appelée lorsqu'un des paramètres du lien est modifié."""

        if not(signal_model == self.signal_model) :
            return

        if not(param_name == 'link_value' or param_name == 'link_type' or param_name == 'text' or param_name == 'color') :
            return

        self._value_semaphore = True

        if param_name == 'link_value' :
            self._sldr_link_value.setValue(int(new_value * sldr_factor))
            self._spn_link_value.setValue(new_value)

        elif param_name == 'link_type' :
            if new_value == LinkType.SQUARE : 
                self._cmb_link_type.setCurrentIndex(0)
            elif new_value == LinkType.CURVED :
                self._cmb_link_type.setCurrentIndex(1)

        elif param_name == 'text' :
            self._fld_texte.setText(new_value)

        elif param_name == 'color' :
            ...

        self._value_semaphore = False

    @pyqtSlot()
    def sldr_link_value_changed (self) -> None :
        """Est la méthode appelée lorsque l'utilisateur modifie la valeur du lien via le slider."""

        if self._value_semaphore :
            return

        if self._link_value_semaphore :
            return

        if self._signal_model is None : 
            return

        self._link_value_semaphore = True
        self._signal_model.action_param_semaphore = True

        new_value = float(self._sldr_link_value.value()) / sldr_factor
        old_value = self.signal_model.link_value
        param_name = 'link_value'

        self.actions_generator(old_value, new_value, param_name)

        self._signal_model.link_value = new_value
        self._spn_link_value.setValue(self._signal_model.link_value)
        self._signal_model.self_update()

        self._signal_model.action_param_semaphore = False
        self._link_value_semaphore = False

    @pyqtSlot()
    def spn_link_value_changed (self) -> None :
        """Est la méthode appelée lorsque l'utilisateur modifie la valeur du lien via le spinner."""

        if self._value_semaphore :
            return

        if self._link_value_semaphore :
            return

        if self._signal_model is None : 
            return

        self._link_value_semaphore = True
        self._signal_model.action_param_semaphore = True

        new_value = self._spn_link_value.value()
        old_value = self.signal_model.link_value
        param_name = 'link_value'

        self.actions_generator(old_value, new_value, param_name)

        self._signal_model.link_value = new_value
        self._sldr_link_value.setValue(int(new_value * sldr_factor))
        self._signal_model.self_update()

        self._signal_model.action_param_semaphore = False
        self._link_value_semaphore = False

    def cmb_link_type_value_changed (self) -> None :
        """Est la méthode appelée lorsque l'utilisateur modifie le type de lien du signal."""

        if self._value_semaphore :
            return

        if self._cmb_link_type.currentIndex() == 0 :
            new_value = LinkType.SQUARE

        elif self._cmb_link_type.currentIndex() == 1 :
            new_value = LinkType.CURVED

        self._signal_model.action_param_semaphore = True
        
        old_value = self.signal_model.link_type
        param_name = 'link_type'

        self.actions_generator(old_value, new_value, param_name)

        self._signal_model.link_type = new_value
        self._signal_model.self_update()

        self._signal_model.action_param_semaphore = False

    def fld_texte_text_changed (self) -> None : 
        """Est la méthode appelée lorsque l'utilisateur modifie le texte du signal."""

        if self._value_semaphore :
            return

        self._signal_model.action_param_semaphore = True

        new_value = self._fld_texte.text()
        old_value = self._signal_model.text
        param_name = 'text'

        self.actions_generator(old_value, new_value, param_name)

        self._signal_model.text = new_value

        self._signal_model.action_param_semaphore = False

    def init_ui (self) :

        if not self.objectName():
            self.setObjectName(u"pnl_signal_properties")
        
        # Texte
        self._lbl_texte = PropertyLabel()
        self._lbl_texte.setObjectName(u"_lbl_texte")

        self._fld_texte = QLineEdit()
        self._fld_texte.setObjectName(u"_fld_texte")
        self._fld_texte.textChanged.connect(self.fld_texte_text_changed)

        row = self.layout().rowCount()
        self.layout().addWidget(self._lbl_texte, row, 0)
        self.layout().addWidget(self._fld_texte, row, 1)

        # Color picker
        self._lbl_color = PropertyLabel()
        self._lbl_color.setObjectName(u"_lbl_color")

        self._cmb_color = QComboBox()
        self._cmb_color.setObjectName(u"_cmb_color")
        self._cmb_color.addItem('black')
        
        row = self.layout().rowCount()
        self.layout().addWidget(self._lbl_color, row, 0)
        self.layout().addWidget(self._cmb_color, row, 1)

        # Link type
        self._lbl_type_of_link = PropertyLabel()
        self._lbl_type_of_link.setObjectName(u"_lbl_type_of_link")

        self._cmb_link_type = QComboBox()
        self._cmb_link_type.setObjectName(u"_cmb_link_type")
        self._cmb_link_type.addItem('Square')
        self._cmb_link_type.addItem('Curved')
        self._cmb_link_type.currentIndexChanged.connect(self.cmb_link_type_value_changed)
        
        row = self.layout().rowCount()
        self.layout().addWidget(self._lbl_type_of_link, row, 0)
        self.layout().addWidget(self._cmb_link_type, row, 1)

        # Link value
        self._lbl_link_value = PropertyLabel()
        self._lbl_link_value.setObjectName(u"_lbl_link_value")

        self._sldr_link_value = QSlider()
        self._sldr_link_value.setObjectName(u"_sldr_link_values")
        self._sldr_link_value.setOrientation(Qt.Horizontal)
        self._sldr_link_value.setMinimum(0)
        self._sldr_link_value.setMaximum(int(sldr_factor))
        self._sldr_link_value.valueChanged.connect(self.sldr_link_value_changed)

        row = self.layout().rowCount()
        self.layout().addWidget(self._lbl_link_value, row, 0)
        self.layout().addWidget(self._sldr_link_value, row, 1)

        self._spn_link_value = DblSpinBox()
        self._spn_link_value.setObjectName(u"_spn_link_value")
        self._spn_link_value.setMinimum(0)
        self._spn_link_value.setMaximum(1)
        self._spn_link_value.setDecimals(2)
        self._spn_link_value.valueChanged.connect(self.spn_link_value_changed)

        row = self.layout().rowCount()
        self.layout().addWidget(PropertyLabel(), row, 0)
        self.layout().addWidget(self._spn_link_value, row, 1)

        # Data type 
        self._lbl_data_type = PropertyLabel()
        self._lbl_data_type.setObjectName(u"_lbl_data_type")

        self._lbl_data_type_value = PropertyLabel()
        self._lbl_data_type_value.setObjectName(u"_lbl_data_type_value")
        
        row = self.layout().rowCount()
        self.layout().addWidget(self._lbl_data_type, row, 0)
        self.layout().addWidget(self._lbl_data_type_value, row, 1)

    def translate_ui (self) -> None :
        self._lbl_texte.setText(QCoreApplication.translate(self.objectName(), u"Texte : ", None))
        self._lbl_color.setText(QCoreApplication.translate(self.objectName(), u"Couleur : ", None))
        self._lbl_type_of_link.setText(QCoreApplication.translate(self.objectName(), u"Type :", None))
        self._lbl_link_value.setText(QCoreApplication.translate(self.objectName(), u"Link value : ", None))
        self._lbl_data_type.setText(QCoreApplication.translate(self.objectName(), u"Data Type : ", None))


#-----------------------------------
# Imports
#-----------------------------------

from typing import List
from math import pow

#-----------------------------------
# Functions
#-----------------------------------

def _function_power_box (box, event_) -> List :
    
    v = pow(box.get_input(0).value, box['power'])
    events : List = []
    
    for output in box.outputs :
        events.append(box.construct_signal_event(output, v))

    return events


#-----------------------------------
# Imports
#-----------------------------------

from typing import List

#-----------------------------------
# Functions
#-----------------------------------

def _init_reccurence_box (box, simulation_params) :

    box['values'] = []
    box['points_count'] = 0

    if len(box.inputs) == 0 :
        box._wait_for_all_signal_events = False
        box._wait_for_events = False
    
    box['values'].append([])

    for x in box.inputs :
        box['values'].append([])

def _function_reccurence_box (box, event_) -> List:
    v = 0

    # Pour chaque entrée ajouter la valeur actuelle dans la liste.
    i = 1
    for input in box.inputs:
        box['values'][i].insert(0, input.value)
        i += 1
    
    box['points_count'] = box['points_count'] + 1

    # Récuperation de la valeur indépenndante
    if len(box['coefficients']) > 0:
        if len(box['coefficients'][0]) > 0 :
            v = box['coefficients'][0][0]
    
    # Anciennes valeurs de sorties
    for j in range(min(len(box['coefficients'][0])-1, len(box['values'][0]))):
        v += box['coefficients'][0][j + 1] * box['values'][0][j]

    # Anciennes valeurs des entrées
    for input_index in range(min((len(box['values']) - 1), (len(box['coefficients']) - 1))) :
        values_index_count = min((len(box['coefficients'][input_index + 1])), (len(box['values'][input_index + 1])))

        for value_index in range(values_index_count) :
            v += (box['coefficients'][input_index + 1][value_index]) * (box['values'][input_index + 1][value_index]) 
    
    # Ajout de la valeur aux anciennes valeurs.
    box['values'][0].insert(0, v)
    
    events : List = []

    for output in box.outputs :
        events.append(box.construct_signal_event(output, v))

    return events


#-----------------------------------
# Imports
#-----------------------------------

from datetime import date
from muphyn.packages.interface.models.editable_models.abstract_editable_model import AbstractEditableModel

#-----------------------------------
# Class
#-----------------------------------

class SchedulerEditorModel (AbstractEditableModel) :
    """Est la classe qui décrit les données contenues dans un éditeur de planificateur."""

    # -------------
    # Constructors
    # -------------

    def __init__(self, name : str, path : str, creator : str, date : date, version : float, code : str = '') :
        
        AbstractEditableModel.__init__(self, name, path, creator, date, version)
        


#-----------------------------------
# Imports
#-----------------------------------

# General Imports
from datetime import date
from typing import Iterable

# Project Imports
from muphyn.packages.interface.models.editable_models.abstract_editable_model import AbstractEditableModel
from muphyn.packages.interface.models.editable_models.abstract_diagram_model import AbstractDiagramModel
from muphyn.packages.interface.models.editable_models.scheduler_model import SchedulerModel
from muphyn.packages.interface.models.graphical_models.abstract_graphical_element import AbstractGraphicalElement

#-----------------------------------
# Class
#-----------------------------------

class SimulationModel (AbstractEditableModel, AbstractDiagramModel) :
    """Est le modèle pour l'éditeur de simulation."""

    # -------------
    # Constructors
    # -------------

    def __init__(self, name : str, path : str, creator : str, date : date, version : float, scheduler_model : SchedulerModel, graphical_elements : Iterable[AbstractGraphicalElement] = []) :
        
        AbstractEditableModel.__init__(self, name, path, creator, date, version)
        AbstractDiagramModel.__init__(self, graphical_elements)
        
        self._scheduler_model : SchedulerModel = scheduler_model

    # -------------
    # Properties
    # -------------

    @property
    def scheduler_model (self) -> SchedulerModel :
        """Permet de récuperer le modèle de planificateur."""
        return self._scheduler_model

    @scheduler_model.setter
    def scheduler_model (self, scheduler_model_ : SchedulerModel) -> None :
        """Permet de modifier le modèle de planificateur."""
        self._scheduler_model = scheduler_model_


#-----------------------------------
# Imports
#-----------------------------------

import math
from typing import Callable, List, Any
from muphyn.packages.core.plci_core_data import Data

from muphyn.packages.core.plci_core_scheduler_exception import SchedulerException

from .plci_core_scheduler_params import SchedulerParams
from .plci_core_scheduler_event import SchedulerEvent
from .plci_core_signal_event import SignalEvent
from .plci_core_signal import Signal
from .plci_core_box import Box
from .plci_core_diagram import Diagram


def round_step_size(quantity, step_size) -> float:
    """Rounds a given quantity to a specific step size
    :param quantity: required
    :param step_size: required
    :return: decimal
    """
    precision: int = int(round(-math.log(step_size, 10), 0))
    return float(round(quantity, precision))

#-----------------------------------
# Class
#-----------------------------------

class Scheduler :
    """Est la classe qui permet de résoudre les simulation basé sur des diagrammes de boxes."""
    
    # -------------
    # Constructors
    # -------------

    def __init__ (self, scheduler_name_ : str, function_ : Callable[[Any], SchedulerException]) :
        
        if function_ is None :
            self._function : Callable[[Any], SchedulerException] = lambda scheduler : ...

        else : 
            self._function = function_

        self._scheduler_name = scheduler_name_
        self._current_time : float = 0
        self._events : List[SignalEvent] = []
        self._diagram : Diagram = Diagram()
        self._scheduler_params : SchedulerParams = SchedulerParams


    # -------------
    # Properties
    # -------------

    @property
    def diagram (self) -> Diagram :
        """Permet de récuperer le diragramme à résoudre."""
        return self._diagram

    @diagram.setter
    def diagram (self, diagram_ : Diagram) -> None :
        """Permet de modifier le diagramme de simulation."""
        self._diagram = diagram_

    @property
    def params (self) -> SchedulerParams :
        """Permet de récuperer les paramètres de la simulation."""
        return self._scheduler_params

    @params.setter
    def params (self, scheduler_params_ : SchedulerParams) -> None :
        """Permet de modifier les paramètres de simulation."""
        self._scheduler_params = scheduler_params_

    @property
    def step_time (self) -> float :
        """Permet de récuperer le temps de step de la simulation."""
        return self._scheduler_params.step_time

    @property
    def stop_time (self) -> float :
        """Permet de récuperer le temps maximal de la simulation."""
        return self._scheduler_params.stop_time

    @property 
    def current_time (self) -> float :
        """Permet de récuperer le temps de simulation acutel."""
        return self._current_time

    @property
    def scheduler_name (self) -> str :
        """Permet de récuperer le nom du planificateur."""
        return self._scheduler_name

    @property 
    def events_count (self) -> int :
        """Permet de récuperer le nombre d'événements dans la liste."""
        return len(self._events)

    @property
    def are_events_left (self) -> bool :
        """Permet de savoir si il reste des événements à réaliser."""
        return len(self._events) > 0

    @property 
    def should_simulation_continue (self) -> bool :
        """Permet de savoir si le temps de simulation n'a pas dépassé encore le temps final de simulation."""
        return self.current_time <= self._scheduler_params.stop_time
    
    # -------------
    # Methods
    # -------------

    def schedule (self, diagram : Diagram = None, params : SchedulerParams = None) -> SchedulerException :
        """Permet de résoudre le diagramme."""

        exc = self._scheduler_call_init()
        if not exc is None :
            return exc
        
        exc = self._function(self)
        if not exc is None :
            return exc
        
        exc = self._scheduler_call_end()
        if not exc is None :
            return exc

        return None

    
    def _scheduler_call_init (self, diagram : Diagram = None) -> SchedulerException :
        """Permet d'appeler la méthode init des boxes contenues dans le diagramme."""
        try :
            for box in self.diagram.boxes :
                box.init(self._scheduler_params)

        except Exception as err :
            return SchedulerException(box, None, None, None, self.diagram, 0, err)

        return None

    def _scheduler_call_end (self) -> SchedulerException :
        """Permet d'appeler la méthode end des boxes contenues dans le diagramme."""
        
        try :
            for box in self.diagram.boxes :
                box.end()

        except Exception as err :
            return SchedulerException(box, None, None, None, self.diagram, self.current_time, err)

        return None

    def stepping_time (self) -> None :
        """Permet d'avancer d'un pas de temps."""
        self._current_time += self._scheduler_params.step_time
        self._current_time = round_step_size(self._current_time, self._scheduler_params.step_time)

    def insert_event (self, position_ : int, signal_event_ : SignalEvent) -> None :
        """Permet d'insérer un évenements dans la liste à un index donné."""
        self._events.insert(position_, signal_event_)

    def append_event (self, signal_event_ : SignalEvent) -> None :
        """Permet d'ajouter un événement à la fin de la liste des événements."""
        self._events.append(signal_event_)

    def remove_event (self, signal_event : SignalEvent) -> None :
        """Permet de supprimer un événement de la liste."""
        self._events.remove(signal_event)
        del signal_event

    def get_event (self, position_ : int) -> SignalEvent :
        """Permet de récuperer un événement suivant son index dans la liste des événements."""
        return self._events[position_]

    def construct_event (self, signal_event_ : SignalEvent = None) -> SchedulerEvent :
        """Permet de créer un schedulerEvent sur base d'un signal event."""
        if signal_event_ == None :
            return SchedulerEvent(None, None, None, self.current_time, self)
        else :
            return SchedulerEvent(signal_event_.signal, signal_event_.box, signal_event_.new_signal_data, self.current_time, self)

    def construct_exception (self, current_box_ : Box, current_box_bis_ : Box, current_event_ : SignalEvent, exception_ : Exception) -> SchedulerException :
        """Permet de créer une exception."""
        return SchedulerException(current_box_, current_box_bis_, self._events, current_event_, self.diagram, self.current_time, exception_)

    def construct_signal_event (self, signal : Signal, box : Box, data : Data) -> SignalEvent :
        """Permet de créer des evenements de signaux."""
        return SignalEvent(signal, box, data)

    def is_scheduler_exception (self, obj : Any) -> bool :
        """Permet de savoir si l'objet passé en paramètre est une exception."""
        return isinstance(obj, SchedulerException)


#-----------------------------------
# Imports
#-----------------------------------

# General Imports
import os
import yaml

# Project Imports
from muphyn.packages.core.utils.log_manager import LogManager
from muphyn.packages.interface.files.abstract_exporter import AbstractExporter
from muphyn.packages.interface.files.abstract_importer import AbstractImporter
from muphyn.packages.interface.files.simulation_files.simulation_exporter import SimulationExporter
from muphyn.packages.interface.files.simulation_files.simulation_importer import SimulationsImporter
from muphyn.packages.interface.models.editable_models.abstract_editable_model import AbstractEditableModel
from muphyn.packages.interface.models.editable_models.box_code_model import BoxCodeModel
from muphyn.packages.interface.models.editable_models.box_composite_model import BoxCompositeModel
from muphyn.packages.interface.models.editable_models.scheduler_model import SchedulerModel 
from muphyn.packages.interface.models.editable_models.simulation_model import SimulationModel
from muphyn.packages.interface.editors.abstract_editor import AbstractEditor

#-----------------------------------
# Functions
#-----------------------------------

def load (path : str) -> AbstractEditableModel :
    """Permet de charger le modèle se trouvant au chemin passé en paramètre."""
    
    importer : AbstractImporter = None
        
    with open(path) as file_data : 

        yaml_data = yaml.load(file_data, Loader = yaml.FullLoader)
        
        name = os.path.basename(path)
        
        path = path[:-name.__len__()]
        if name.endswith('.yaml') :
            name = name[:-('.yaml'.__len__())]

        if 'simulation' in yaml_data :
            importer = SimulationsImporter()
            data = yaml_data['simulation']

        if importer is None : 
            return None

        return importer.open(data, path, name)

def save (model : AbstractEditableModel, path : str) -> bool :
    """Permet de sauvegarder le modèle au chemin passé en paramètre."""
    
    exporter : AbstractExporter = None

    if isinstance(model, SimulationModel):
        exporter = SimulationExporter()

    elif isinstance(model, BoxCompositeModel) :
        LogManager().debug('Save box composite !')
        raise Exception('no exporter found for box composite model')

    elif isinstance(model, SchedulerModel) :
        LogManager().debug('Save scheduler !')
        raise Exception('no exporter found for scheduler model')

    elif isinstance(model, BoxCodeModel) :
        LogManager().debug('Save box code !')
        raise Exception('no exporter found for box code model')

    return exporter.save(model, path)

def export (editor : AbstractEditor, argument : str) :
    """Permet d'exporter l'éditeur sous une forme voulue."""


#-----------------------------------
# Imports
#-----------------------------------

from typing import List, Dict, Any
from muphyn.packages.interface.graphical_actions.abstract_diagram_action import AbstractDiagramAction
from muphyn.packages.interface.models.graphical_models.box_model import BoxModel
from muphyn.packages.interface.models.signals_model.abstract_signal_model import AbstractSignalModel

#-----------------------------------
# Class
#-----------------------------------

class DiagramAddGraphicalElementAction (AbstractDiagramAction) :
    """Est le type d'action qui permet d'ajouter des éléments dans l'écran."""

    # -------------
    # Constructors
    # -------------

    def __init__ (self, diagram_model : Any, constructors_ : List[Dict]) :
        
        AbstractDiagramAction.__init__(self, diagram_model)

        self._elements_indices = []
        self._constructors = constructors_
    
    # -------------
    # Methods
    # -------------

    def do (self) :
        
        for constructor in self._constructors :
            # Build Box Model
            box_model = BoxModel.fromBoxData(constructor['box_data'], position=constructor['pos'])

            # Add Box Model to diagram
            self.diagram_model.add_element(box_model)

            box_model_inputs = list(box_model.inputs)
            box_model_outputs = list(box_model.outputs)

            if 'graphical_index' in constructor :
                box_model.graphical_index = constructor['graphical_index']

                for input_index, input in enumerate(constructor['inputs']) :
                    box_model_inputs[input_index].graphical_index = input

                for output_index, output in enumerate(constructor['outputs']) :
                    box_model_outputs[output_index].graphical_index = output

            else :
                constructor['graphical_index'] = box_model.graphical_index
                constructor['inputs'] = []
                constructor['outputs'] = []

                for input in box_model.inputs : 
                    constructor['inputs'].append(input.graphical_index)

                for output in box_model.outputs : 
                    constructor['outputs'].append(output.graphical_index)

            self._elements_indices.append(box_model.graphical_index)

    def undo (self) :
        
        for graphical_element in self.diagram_model._graphical_elements :

            if graphical_element.graphical_index in self._elements_indices :

                if isinstance(graphical_element, AbstractSignalModel) :
                    graphical_element.unbind()

                self.diagram_model.remove_element(graphical_element)


            if len(self._elements_indices) == 0 : 
                break


# General imports
import numbers
from typing import Any

class AbstractSimulation:
    def __init__(self) -> None:

        # Model Informations
        self._inputs: dict = {}
        self._outputs: dict = {}
        self._parameters: dict = {}
        self._components: dict = {}

        # Simulation Informations
        self._start_time: float = None
        self._stop_time: float = None
        self._step_size: float = None

    # -------------
    # Properties
    # -------------
    @property
    def inputs(self) -> dict:
        return self._inputs

    @property
    def outputs(self) -> dict:
        return self._outputs

    @property
    def parameters(self) -> dict:
        return self._parameters

    @property
    def components(self) -> dict:
        return self._components

    @property
    def start_time(self) -> numbers.Number:
        return self._start_time

    @start_time.setter
    def start_time(self, new_start_time: numbers.Number) -> None:
        if AbstractSimulation.is_valid_start_time(new_start_time) and new_start_time != self._start_time:
            self._start_time = new_start_time

    @property
    def stop_time(self) -> numbers.Number:
        return self._stop_time

    @stop_time.setter
    def stop_time(self, new_stop_time: numbers.Number) -> None:
        if AbstractSimulation.is_valid_stop_time(new_stop_time) and new_stop_time != self._stop_time:
            self._stop_time = new_stop_time

    @property
    def step_size(self) -> numbers.Number:
        return self._step_size

    @step_size.setter
    def step_size(self, new_step_size: numbers.Number) -> None:
        if AbstractSimulation.is_valid_step_size(new_step_size) and new_step_size != self._step_size:
            self._step_size = new_step_size

    # -------------
    # Methods
    # -------------
    def run_simulation(self):
        raise(NotImplementedError(f"{type(self).__name__}.run_simulation() not implemented yet"))

    def set_simulation_options(self, start_time: float, stop_time: float, step_size:float):
        if AbstractSimulation.is_valid_start_time(start_time):
            self._start_time = start_time

        if  AbstractSimulation.is_valid_stop_time(stop_time):
            self._stop_time = stop_time

        if  AbstractSimulation.is_valid_step_size(step_size):
            self._step_size = step_size

    # --------------
    # Static Methods
    # --------------
    @staticmethod
    def is_valid_start_time(start_time: Any):
        return start_time is not None and isinstance(start_time, numbers.Number)

    @staticmethod
    def is_valid_stop_time(stop_time: Any):
        return stop_time is not None and isinstance(stop_time, numbers.Number)

    @staticmethod
    def is_valid_step_size(step_size: Any):
        return step_size is not None and isinstance(step_size, numbers.Number)



#-----------------------------------
# Imports
#-----------------------------------

# General Imports
import os
from typing import Union

# PyQt5 Imports
from PyQt5.QtCore import QPointF, QSize, QSizeF, Qt
from PyQt5.QtWidgets import QGraphicsItem

# Project Imports
from muphyn.packages.interface.shapes.basic_shape import GroupedShapes, Text, ImageShapeFactory, AbstractImage
from muphyn.packages.interface.utils.interface_constants import MuphynFonts

class Icon(GroupedShapes):
    def __init__(self, iconPath: str, description: str = None, position: QPointF = QPointF(), 
            size: QSizeF = None, parent: QGraphicsItem = None) -> None:

        # Init parent class
        super().__init__(position, parent)

        # Icon path
        self.iconPath = iconPath

        # Test if path exists
        if os.path.exists(iconPath):
            # Build icon shape
            self.imageShape: AbstractImage = ImageShapeFactory(image_path=iconPath, position=position, size=size, parent=self)

        else:
            # Handle description
            if description is None : 
                self.description = os.path.basename(iconPath)

            # Build text flag
            alignementFlag = Qt.AlignmentFlag.AlignVCenter

            # Build Text shape
            self.textShape = Text(self.description, position=QPointF(), 
                font=MuphynFonts.BoxModelDetailsFont, alignment=alignementFlag, parent=self)

    def setWidth(self, new_width: int) -> None:
        if hasattr(self, "imageShape"):
            self.imageShape.setWidth(new_width)
        else:
            # Resize text area
            pass

    def setHeight(self, new_height: QSizeF) -> None:
        if hasattr(self, "imageShape"):
            self.imageShape.setHeight(new_height)
        else:
            # Resize text area
            pass

    def setSize(self, new_size: Union[QSize, QSizeF]) -> None:
        if hasattr(self, "imageShape"):
            self.imageShape.setSize(new_size)
        else:
            # Resize text area
            pass






from muphyn.packages.interface.utils.graphical_items.base_items.icon import Icon
from muphyn.packages.interface.utils.graphical_items.box_model_area import BoxModelInfoArea
from muphyn.packages.interface.utils.graphical_items.file_selector_button import AnyFileSelectorButton, DirectorySelectorButton, ExistingFileSelectorButton, ExistingFilesSelectorButton, file_selector_button



# Integer
IntegerRegex = r"^[-+]?[0-9]*$"

# Float with comma as decimal separator
CommaFloatRegex = r"^[-+]?[0-9]*,[0-9]*$"

# Float with point as decimal separator
DotFloatRegex = r"^[-+]?[0-9]*(?:\.[0-9]*)?$"

# Float with comma as decimal separator (Scientific Notation)
CommaScientificNumberRegex = r"^-?[1-9](?:,\d+)?[Ee][-+]?\d+$"

# Float with point as decimal separator (Scientific Notation)
DotScientificNumberRegex = r"^-?[1-9](?:\.\d+)?[Ee][-+]?\d+$"



from PyQt5.QtWidgets import QAction
from PyQt5.QtGui import QIcon
from os.path import dirname
from muphyn.utils.paths import ROOT_DIR

class StartAction(QAction):
    def __init__(self, text: str = "Start", parent=None):
        super().__init__(QIcon(ROOT_DIR + "/" + "packages/interface/assets/GeneralIcons/start.svg"), text, parent)

class StopAction(QAction):
    def __init__(self, text: str = "Stop", parent=None):
        super().__init__(QIcon(ROOT_DIR + "/" + "packages/interface/assets/GeneralIcons/stop.svg"), text, parent)

class RotateLeftAction(QAction):
    def __init__(self, text: str = "Rotate 90° Left", parent=None):
        super().__init__(QIcon(ROOT_DIR + "/" + "packages/interface/assets/GeneralIcons/rotate_left.svg"), text, parent)

class RotateRightAction(QAction):
    def __init__(self, text: str = "Rotate 90° Right", parent=None):
        super().__init__(QIcon(ROOT_DIR + "/" + "packages/interface/assets/GeneralIcons/rotate_right.svg"), text, parent)

class UndoAction(QAction):
    def __init__(self, text: str = "Rotate 90° Right", parent=None):
        super().__init__(QIcon(ROOT_DIR + "/" + "packages/interface/assets/GeneralIcons/undo.svg"), text, parent)

class RedoAction(QAction):
    def __init__(self, text: str = "Rotate 90° Right", parent=None):
        super().__init__(QIcon(ROOT_DIR + "/" + "packages/interface/assets/GeneralIcons/redo.svg"), text, parent)

class DeleteAction(QAction):
    def __init__(self, text: str = "Delete", parent=None):
        super().__init__(QIcon(ROOT_DIR + "/" + "packages/interface/assets/GeneralIcons/delete.svg"), text, parent)


from os import system
import matplotlib.pyplot as plt

from typing import List

from box_library.transfert_function_box import get_coeff_vector, euler, trapeze

_ = system('cls')

# =========== Differential methods ===========


def differentiate(order: int, last_index: int, vector: List[float], last_derivatives: List[float], dt: float) -> List[float]:

    returning_tuple = []

    if last_index > 0:
        d_1 = vector[last_index]
        d_2 = vector[last_index - 1]
        returning_tuple.append(euler(d_1, d_2, dt))

        for current_order in range(max(min(order, last_index - 1), 0)):
            d_1 = returning_tuple[current_order]
            d_0 = last_derivatives[last_index - 1][current_order]
            current_derivative = euler(d_1, d_0, dt)

            returning_tuple.append(current_derivative)

            if current_derivative == 0:
                break

    while len(returning_tuple) < order:
        returning_tuple.append(0)

    return returning_tuple

# =========== Differential equation ===========

num = '1 1 0'

num_vect = get_coeff_vector(num)
num_vect.reverse()

num_order = len(num_vect) - 1

# =========== Simulation ===========

fs = 1000
dt = 1/fs
sim_time = 20
samples_number = int(fs * sim_time) + 1

print('=============================================================')
print('||                                                         ||')
print('||                      Simulation                         ||')
print('||                                                         ||')
print('|| X = Y * (P² + P)                                        ||')
print('||                                                         ||')
print('|| - num order :', num_order, '                                        ||')
print('||                                                         ||')
print('|| - simulation time :', sim_time, '                                 ||')
print('|| - sample frequency :', fs, '                              ||')
print('|| - dt :', dt, '                                           ||')
print('|| - samples number :', samples_number, '                               ||')
print('||                                                         ||')
print('=============================================================')
print('')
print('')

# time
t = [n * dt for n in range(samples_number)]

# input
u = [(12.5*t_)**2 for t_ in t]

# output
u_derivatives : List[List[float]] = []
y = []


for i, t_ in enumerate(t) :
#for i in range(10) :

    current_y = 0

    current_u_derivative = differentiate(num_order, i, u, u_derivatives, dt)

    for j, current_coeff in enumerate(num_vect):

        if j == 0:
            current_y += (current_coeff * u[i])

        else:
            current_y += (current_coeff * current_u_derivative[j - 1])

    u_derivatives.append(current_u_derivative)

    y.append(current_y)


print('at', (len(y) - 1) / fs, 's :', y[len(y) - 1])

# =========== Personnal notes ===========

print('')
print(' =====> Résultats <=====')
print('')
print('date : 16/02/2022')
print('')
print('On se rend bien compte que ma méthode fonctionne pour dériver des u.')
print('Cependant, ma méthode d\'approche pour les y semble très mauvaise !')
print('')
print('')
print('date : 17/02/2022')
print('')
print('Essai d\'intégration des U au lieu de dériver les Y ... TRES MAUVAISES REPONSES !!!')
print('')
print('')
print('date : 18/02/2022')
print('')
print('Reprise depuis le début du travail, voir EDT_SFT_0048_3 pour voir la suite.')

# =========== Plot results ===========

t_renderable = t[0 : len(y)]
u_renderable = u[0 : len(y)]

plt.figure()
plt.plot(t_renderable, u_renderable, label="u(t)")
plt.plot(t_renderable, y, label="y(t)")
plt.legend()
plt.grid()
plt.xlabel('Time')
plt.show()




from muphyn.packages.core.plci_core_scheduler_params import SchedulerParams
from ..core.plci_core_scheduler_event import SchedulerEvent
from ..core.plci_core_box import Box

def box_init_method (box : Box) -> None :
    print("init method")

def box_function_method (box : Box, scheduler_event : SchedulerEvent) -> None :
    print("function method")

def box_end_method (box : Box) -> None:
    print("end method")

b = Box(0, 'a', 'b', box_init_method, box_function_method, box_end_method)

b.init(SchedulerParams())
b.function(SchedulerEvent(None, None, None, 0, None))
b.end()


#-----------------------------------
# Imports
#-----------------------------------

import sys
import math as m
from PyQt5 import QtGui

from PyQt5.QtWidgets import QDoubleSpinBox, QWidget

#-----------------------------------
# Class
#-----------------------------------

class DblSpinBox (QDoubleSpinBox) :
    """Est la classe qui décrit une spin box gérant des doubles dont le nombre de zéro est variable."""

    # -------------
    # Contructors
    # -------------

    def __init__ (self, parent : QWidget = None) :

        QDoubleSpinBox.__init__(self, parent)
    
        self.setMinimum(-sys.float_info.max)
        self.setMaximum(sys.float_info.max)
        self.setValue(0)
        self.setDecimals(10) 
    
    # -------------
    # Methods
    # -------------

    def keyPressEvent(self, event : QtGui.QKeyEvent) -> None :
        return super().keyPressEvent(event)

    def setValue (self, value : float) -> None :
        return super().setValue(value)

    def setSingleStep (self, val : float) -> None :
        ...

    def setDecimals (self, precision : int) -> None :

        if precision < 0 :
            precision = 0

        super().setSingleStep(10 * m.pow(10, -precision))
        return super().setDecimals(precision) 




from ..core.plci_core_scheduler import Scheduler
from ..core.plci_core_data_type import DataType
from ..core.plci_core_signal import Signal
from ..core.plci_core_scheduler_params import SchedulerParams

from .unit_test import get_boxes_libraries, get_schedulers_library, simulate

# Get library
boxesLibraries = get_boxes_libraries()
schedulersLibraries = get_schedulers_library()

# Scheduler parameters
scheduler : Scheduler = schedulersLibraries.construct_scheduler('Schedulers', 'Default')
scheduler.params = SchedulerParams(4.0, 0.001)

scheduler.diagram.append(Signal(index_ = 0, signal_type_ = DataType.FLOAT, default_value_ = 0.0))
scheduler.diagram.append(Signal(index_ = 1, signal_type_ = DataType.FLOAT, default_value_ = 0.0))
scheduler.diagram.append(Signal(index_ = 2, signal_type_ = DataType.FLOAT, default_value_ = 0.0))
scheduler.diagram.append(Signal(index_ = 3, signal_type_ = DataType.FLOAT, default_value_ = 0.0))
scheduler.diagram.append(Signal(index_ = 4, signal_type_ = DataType.FLOAT, default_value_ = 0.0))
scheduler.diagram.append(Signal(index_ = 5, signal_type_ = DataType.FLOAT, default_value_ = 0.0))

scheduler.diagram.append(boxesLibraries.construct_box('Boxes.Sources', 'Step', start_value = 0, stop_value = 1, step_time = 1))
scheduler.diagram.append(boxesLibraries.construct_box('Boxes.Math', 'Addition'))
scheduler.diagram.append(boxesLibraries.construct_box('Boxes.Sources', 'Constant', value = 3))
scheduler.diagram.append(boxesLibraries.construct_box('Boxes.Math', 'Multiplier'))
scheduler.diagram.append(boxesLibraries.construct_box('Boxes.Math', 'Integrator'))
scheduler.diagram.append(boxesLibraries.construct_box('Boxes.Signals', 'Amplifier', gain = -0.98))
scheduler.diagram.append(boxesLibraries.construct_box('Boxes.Sinks', 'Graph', start_time = 0, stop_time = scheduler.stop_time, title = "step", label_y = "Amplitude (V)", label_x = "Time (s)"))
scheduler.diagram.append(boxesLibraries.construct_box('Boxes.Sinks', 'Graph', start_time = 0, stop_time = scheduler.stop_time, title = "output", label_y = "Amplitude (V)", label_x = "Time (s)"))
 
# Diagram creation
#ODE

scheduler.diagram.add_box_outputs(scheduler.diagram.boxes[0], scheduler.diagram.signals[0])
scheduler.diagram.add_box_outputs(scheduler.diagram.boxes[1], scheduler.diagram.signals[1])
scheduler.diagram.add_box_outputs(scheduler.diagram.boxes[2], scheduler.diagram.signals[2])
scheduler.diagram.add_box_outputs(scheduler.diagram.boxes[3], scheduler.diagram.signals[3])
scheduler.diagram.add_box_outputs(scheduler.diagram.boxes[4], scheduler.diagram.signals[4])
scheduler.diagram.add_box_outputs(scheduler.diagram.boxes[5], scheduler.diagram.signals[5])

scheduler.diagram.add_box_inputs(scheduler.diagram.boxes[1], scheduler.diagram.signals[0], scheduler.diagram.signals[5])
scheduler.diagram.add_box_inputs(scheduler.diagram.boxes[3], scheduler.diagram.signals[1], scheduler.diagram.signals[2])
scheduler.diagram.add_box_inputs(scheduler.diagram.boxes[4], scheduler.diagram.signals[3])
scheduler.diagram.add_box_inputs(scheduler.diagram.boxes[5], scheduler.diagram.signals[4])
scheduler.diagram.add_box_inputs(scheduler.diagram.boxes[6], scheduler.diagram.signals[0])
scheduler.diagram.add_box_inputs(scheduler.diagram.boxes[7], scheduler.diagram.signals[4])

# Launch solver
if simulate(scheduler) :
    print("===============================================")
    print("Values : ")
    l = len(scheduler.diagram.boxes[7].params['data_y'])
    for i in range(l) :
        print(str(scheduler.diagram.boxes[7].params['data_y'][i]).replace('.', ','))


# General import
from enum import Enum

# PyQt import
from PyQt5.QtCore import QPointF
from PyQt5.QtGui import QColor
from PyQt5.QtWidgets import QGraphicsItem

# Project import
from muphyn.packages.interface.editors.multiphysics_model import AbstractComponentModel, ComponentConnector
from muphyn.packages.interface.shapes.basic_shape.shapes import Circle

class ElectricalComponentConnector(ComponentConnector):

    class ConnectorType(Enum):
        Positive = 1
        Negative = -1
        Neutral = 0

    # Constants
    Radius = ComponentConnector.ItemSideSize / 2

    ConnectorFillColorLUT = {
        ConnectorType.Positive: QColor(50, 50, 50),
        ConnectorType.Negative: QColor(255, 255, 255),
        ConnectorType.Neutral: QColor(128, 128, 128)
    }

    def __init__(self, connectorType: ConnectorType, group_position: QPointF = QPointF(), parent: QGraphicsItem = None) -> None:
        super().__init__(group_position, parent)

        # Select border & fill colors
        fillColor = ElectricalComponentConnector.ConnectorFillColorLUT[connectorType]

        # Circle
        center = group_position + QPointF(ElectricalComponentConnector.Radius, ElectricalComponentConnector.Radius)
        radius = ElectricalComponentConnector.Radius
        Circle(center, radius, fill_color=fillColor, border_width=2, parent=parent)

# 
class GroundModel(AbstractComponentModel):

    # Constants
    IconPath = "packages/interface/assets/Components/Electrics/ground.svg"

    def __init__(self, group_position = QPointF(), parent = None) -> None:
        super().__init__(__class__.IconPath, group_position, parent)

        # Calculate node position
        pos = QPointF((AbstractComponentModel.DefaultSize.width())/2, 0) - QPointF(ElectricalComponentConnector.Radius, ElectricalComponentConnector.Radius)

        self.p: ComponentConnector = ElectricalComponentConnector(
            ElectricalComponentConnector.ConnectorType.Neutral, 
            pos, 
            parent=self
        )

# Resistor Models
class ResistorModel(AbstractComponentModel):

    # Constants
    IconPath = "packages/interface/assets/Components/Electrics/resistor.svg"

    def __init__(self, group_position = QPointF(), parent = None) -> None:
        super().__init__(__class__.IconPath, group_position, parent)

        # Calculate point position
        leftX =  -ElectricalComponentConnector.Radius
        rightX = AbstractComponentModel.DefaultSize.width() - ElectricalComponentConnector.Radius
        y = AbstractComponentModel.DefaultSize.height()/2 - ElectricalComponentConnector.Radius

        # Add connectors
        self.p: ComponentConnector = ElectricalComponentConnector(
            ElectricalComponentConnector.ConnectorType.Positive, 
            QPointF(leftX, y), 
            parent=self
        )
        self.n: ComponentConnector = ElectricalComponentConnector(
            ElectricalComponentConnector.ConnectorType.Negative, 
            QPointF(rightX, y), 
            parent=self
        )

class VariableResistorModel(AbstractComponentModel):

    # Constants
    IconPath = "packages/interface/assets/Components/Electrics/variable_resistor.svg"

    def __init__(self, group_position = QPointF(), parent = None) -> None:
        super().__init__(__class__.IconPath, group_position, parent)

        # Calculate point position
        leftX =  -ElectricalComponentConnector.Radius
        rightX = AbstractComponentModel.DefaultSize.width() - ElectricalComponentConnector.Radius
        y = AbstractComponentModel.DefaultSize.height()/2 - ElectricalComponentConnector.Radius

        # Add connectors
        self.p: ComponentConnector = ElectricalComponentConnector(
            ElectricalComponentConnector.ConnectorType.Positive, 
            QPointF(leftX, y), 
            parent=self
        )
        self.n: ComponentConnector = ElectricalComponentConnector(
            ElectricalComponentConnector.ConnectorType.Negative, 
            QPointF(rightX, y), 
            parent=self
        )

class SignalCurrentSourceModel(AbstractComponentModel):

    # Constants
    IconPath = "packages/interface/assets/Components/Electrics/signal_current_source.svg"

    def __init__(self, group_position = QPointF(), parent = None) -> None:
        super().__init__(__class__.IconPath, group_position, parent)

        # Calculate point position
        leftX =  -ElectricalComponentConnector.Radius
        rightX = AbstractComponentModel.DefaultSize.width() - ElectricalComponentConnector.Radius
        y = AbstractComponentModel.DefaultSize.height()/2 - ElectricalComponentConnector.Radius

        # Add connectors
        self.p: ComponentConnector = ElectricalComponentConnector(
            ElectricalComponentConnector.ConnectorType.Positive, 
            QPointF(leftX, y), 
            parent=self
        )

        self.n: ComponentConnector = ElectricalComponentConnector(
            ElectricalComponentConnector.ConnectorType.Negative, 
            QPointF(rightX, y), 
            parent=self
        )



#-----------------------------------
# Imports
#-----------------------------------

from PyQt5.QtCore import QCoreApplication
from PyQt5.QtWidgets import QLabel
from muphyn.packages.interface.models.graphical_models.box_model import BoxModel
from muphyn.packages.interface.properties_pages.abstract_properties_editor import AbstractPropertiesEditor
from muphyn.packages.interface.widgets.BaseWidgets.Label import PropertyLabel

#-----------------------------------
# Class
#-----------------------------------

class BoxProperties (AbstractPropertiesEditor) :
    """Est la classe qui affiche une page capable de modifier les propriétés d'une box."""

    # -------------
    # Constructors
    # -------------

    def __init__ (self, box_model : BoxModel) :

        AbstractPropertiesEditor.__init__(self, box_model)

        self._box_model : BoxModel = None
        self.box_model = box_model

    # -------------
    # Properties
    # -------------

    @property
    def box_model (self) -> BoxModel :
        """Permet de récuperer le box model dont les propriétées sont changées."""
        return self._box_model

    @box_model.setter
    def box_model (self, box_model_ : BoxModel) -> None :
        """Permet de modifier le box model dont les propriétées sont changées."""

        self._box_model = box_model_
        self._model = box_model_
        self.moveable_element = box_model_

        if self._box_model is None : 
            self._lbl_library_value.setText('')
            self._lbl_name_value.setText('')

        else :
            self._lbl_library_value.setText(self._box_model.library)
            self._lbl_name_value.setText(self._box_model.name)

    # -------------
    # Methods
    # -------------

    def init_ui (self) :
        if not self.objectName() :
            self.setObjectName(u"pnl_box_properties")

        self._lbl_library : PropertyLabel = PropertyLabel()
        self._lbl_library.setObjectName(u'_lbl_library')

        self._lbl_library_value : PropertyLabel = PropertyLabel()
        self._lbl_library_value.setObjectName(u'_lbl_library_value')

        # Adding Library name
        row = self.layout().rowCount()
        self.layout().addWidget(self._lbl_library, row, 0)
        self.layout().addWidget(self._lbl_library_value, row, 1)

        self._lbl_name : PropertyLabel = PropertyLabel()
        self._lbl_name.setObjectName(u'_lbl_name')

        self._lbl_name_value : PropertyLabel = PropertyLabel()
        self._lbl_name_value.setObjectName(u'_lbl_name_value')

        # Adding Box name
        row = self.layout().rowCount()
        self.layout().addWidget(self._lbl_name, row, 0)
        self.layout().addWidget(self._lbl_name_value, row, 1)

    def translate_ui (self) -> None :
        
        self._lbl_library.setText(QCoreApplication.translate(self.objectName(), u"Library : ", None))
        self._lbl_name.setText(QCoreApplication.translate(self.objectName(), u"Name : ", None))



from PyQt5.QtGui import QBrush, QColor, QPen
from PyQt5.QtCore import Qt

# -------------
# Pens
# -------------
def SelectedPen(color: QColor) -> QPen:
    selected_pen : QPen = QPen(color)
    selected_pen.setWidth(1)
    selected_pen.setDashPattern([3, 2])
    return selected_pen

def UnselectedPen(color: QColor) -> QPen:
    unselected_pen : QPen = QPen(color)
    unselected_pen.setWidth(1)
    return unselected_pen

# Black Pens
SelectedBlackPen: QPen = SelectedPen(Qt.GlobalColor.black)
UnselectedBlackPen: QPen = UnselectedPen(Qt.GlobalColor.black)

# Red Pens
SelectedRedPen : QPen = SelectedPen(Qt.GlobalColor.red)
UnSelectedRedPen : QPen = UnselectedPen(Qt.GlobalColor.red)

# -------------
# Brushes
# -------------
WhiteBrush : QBrush = QBrush(Qt.GlobalColor.white)
BlackBrush : QBrush = QBrush(Qt.GlobalColor.black)



from typing import List, Any

from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtCore import QPointF, QSizeF, Qt
from PyQt5.QtWidgets import QDialog, QGraphicsItem, QGraphicsScene, QGraphicsView, QStyleOptionGraphicsItem, QWidget

from muphyn.packages.core.plci_core_data_type import DataType
from muphyn.packages.interface.models.editable_models.abstract_diagram_model import AbstractDiagramModel
from muphyn.packages.interface.models.graphical_models.abstract_box_model import AbstractBoxModel
from muphyn.packages.interface.models.graphical_models.abstract_graphical_element import AbstractGraphicalElement
from muphyn.packages.interface.models.graphical_models.box_model import BoxModel

class SelectableBox (AbstractGraphicalElement) :

    def __init__ (self, name : str, position : QPointF, size : QSizeF, parent : QGraphicsItem = None) :
        
        AbstractGraphicalElement.__init__(self, name, position, size, 0, name, parent)
        
        self.setFlag(QGraphicsItem.ItemIsSelectable)
        self.setFlag(QGraphicsItem.ItemIsMovable)
        self.setFlag(QGraphicsItem.ItemSendsScenePositionChanges)

        self._font_metrics_used : QtGui.QFontMetrics = None
        self._text_point : QPointF = QPointF(0, 0)

    def itemChange(self, change: QGraphicsItem.GraphicsItemChange, value: Any) -> Any :

        if change == QGraphicsItem.ItemSelectedChange :
            self.setZValue(value)

        elif change == QGraphicsItem.ItemPositionHasChanged :
            print('item changed ', self.name, ' position : ', value.x(), '; ', value.y())

        elif change == QGraphicsItem.ItemPositionChange :
            print('item changed ', self.name, ' position : ', value.x(), '; ', value.y())

        return super().itemChange(change, value)


    def recompute_name_position (self) :
        """Permet de recalculer la taille du texte affiché et le recentrer dans la box."""
        
        if self._font_metrics_used is None :
            return

        point_x = (self.size.width() - self._font_metrics_used.width(self.rendered_text)) / 2
        point_y = ((self.size.height() - self._font_metrics_used.height()) / 2) + self._font_metrics_used.height()
        self._text_point = QPointF(point_x, point_y)

    def paint(self, painter: QtGui.QPainter, option: QStyleOptionGraphicsItem, widget) -> None :

        if self._font_metrics_used is None :
            self._font_metrics_used = QtGui.QFontMetrics(painter.font())

        self.recompute_name_position()

        if self.isSelected() :
            painter.setPen(self.selected_black_pen)
        
        else :
            painter.setPen(self.unselected_black_pen)
        
        painter.fillRect(self._bound, self.white_brush)
        painter.drawRect(self._bound)
        painter.drawText(int(self._text_point.x()), int(self._text_point.y()), self.rendered_text)


class Scenery (QGraphicsScene) :

    def __init__ (self, parent : QWidget = None, graphical_scenery = None) :
        QGraphicsScene.__init__(self, parent)
        self._graphical_scenery = graphical_scenery
        self._start_sliding_point = None
        self._boxes : List[AbstractGraphicalElement] = []

    def append_element (self, box : AbstractBoxModel) -> None :
        self.addItem(box)
        self._boxes.append(box)

    def new_selection_item (self, selectable) -> None :
        self.setFocusItem(selectable)
        selectable.setZValue(self.items().__len__())

class GraphicalScenery (QGraphicsView) :

    def __init__ (self, parent : QWidget) :

        QGraphicsView.__init__(self, parent)

        self.setDragMode(QGraphicsView.ScrollHandDrag)
        self.setDragMode(QGraphicsView.RubberBandDrag)

    def mousePressEvent(self, event):
        if event.button() == Qt.MidButton: # or Qt.MiddleButton
            self.__prevMousePos = event.pos()
            event.accept()
        else:
            super().mousePressEvent(event)

    def mouseMoveEvent (self, event) :
        if event.buttons() == Qt.MidButton: # or Qt.MiddleButton
            offset = self.__prevMousePos - event.pos()
            self.__prevMousePos = event.pos()

            self.verticalScrollBar().setValue(self.verticalScrollBar().value() + offset.y())
            self.horizontalScrollBar().setValue(self.horizontalScrollBar().value() + offset.x())
            event.accept()
        else:
            super().mouseMoveEvent(event)

    def wheelEvent(self, event: QtGui.QWheelEvent) -> None:

        if event.modifiers() == Qt.ControlModifier :

            if event.angleDelta().y() > 0 :
                self.scale(1.125, 1.125)

            elif event.angleDelta().y() < 0 :
                self.scale(0.825, 0.825)

            event.accept()
        else :
            super().wheelEvent(event) 

class UI (QDialog) :

    def __init__ (self, parent : QWidget = None) :

        QDialog.__init__(self, parent)
        self.setFixedSize(630, 480)
        self.setWindowTitle('Test graphics')


        self._red_brush : QtGui.QBrush = QtGui.QBrush(Qt.red)
        self._blue_brush : QtGui.QBrush = QtGui.QBrush(Qt.blue)
        self._black_pen : QtGui.QPen = QtGui.QPen(Qt.black)
        self._black_pen.setWidth(6)

        
        self._graphics_view : QGraphicsView = AbstractDiagramModel(self)
        self._graphics_view.setGeometry(QtCore.QRect(10, 10, 610, 460))
        
        width = 100
        height = 150

        box1 = BoxModel('sources', 'box 1', QPointF(10, 10), QSizeF(width, height), 0, True, parent = None)
        box1.insert_input(0, 'Test', DataType.FLOAT, 'Test')
        box1.insert_input(1, 'Test', DataType.FLOAT, 'Test')
        box1.insert_output(0, 'Test', DataType.FLOAT, 'test')
        box1.insert_output(2, 'Test', DataType.FLOAT, 'test')

        box2 = BoxModel('sources', 'box 2', QPointF(20 + width + width, 10), QSizeF(width, height), 0, True, parent = None)
        box2.insert_input(0, 'Test', DataType.FLOAT, 'Test')
        box2.insert_output(0, 'Test', DataType.FLOAT, 'test')

        box6 = BoxModel('sources', 'box 6', QPointF(20 + width + width, 20 + height), QSizeF(width, height), 0, True, parent = None)
        box6.insert_input(0, 'Test', DataType.FLOAT, 'Test')
        box6.insert_output(0, 'Test', DataType.FLOAT, 'test')

        box3 = BoxModel('sources', 'box 3', QPointF(10, 20 + height), QSizeF(width, height), 90, True, parent = None)
        box4 = BoxModel('sources', 'box 4', QPointF(10, 30 + height + height), QSizeF(width, height), 180, True, parent = None)
        box5 = BoxModel('sources', 'box 5', QPointF(10, 40 + height + height + height), QSizeF(width, height), 270, True, parent = None)

        box3.insert_input(0, 'Test', DataType.FLOAT, 'Test')
        box3.insert_output(0, 'Test', DataType.FLOAT, 'test')

        box4.insert_input(0, 'Test', DataType.FLOAT, 'Test')
        box4.insert_output(0, 'Test', DataType.FLOAT, 'test')

        box5.insert_input(0, 'Test', DataType.FLOAT, 'Test')
        box5.insert_output(0, 'Test', DataType.FLOAT, 'test')

        self._graphics_view.add_graphical_element(box1)
        self._graphics_view.add_graphical_element(box2)
        #self._graphics_view.add_graphical_element(box3)
        #self._graphics_view.add_graphical_element(box4)
        #self._graphics_view.add_graphical_element(box5)
        self._graphics_view.add_graphical_element(box6)


if __name__ == "__main__":
    import sys

    app = QtWidgets.QApplication(sys.argv)
    ui = UI()
    ui.show()
    sys.exit(app.exec_())


from PyQt5.QtWidgets import QMessageBox, QPushButton

from muphyn.packages.interface.widgets.BaseWidgets.Button import PlainButton
from muphyn.packages.interface.utils.interface_constants import ApplicationWindowTitle

class YesNoMessageBox(QMessageBox):

    def __init__(self, question: str, title: str=ApplicationWindowTitle, parent = None):
        super().__init__(parent)

        # Set Question Text
        self.setText(question)

        # Set Title
        self.setWindowTitle(title)

        # Add buttons
        self.addButton(PlainButton(text="Oui"), QMessageBox.ButtonRole.AcceptRole)
        self.addButton(QPushButton(text="Non"), QMessageBox.ButtonRole.RejectRole)




from ..core.plci_core_scheduler import Scheduler
from ..core.plci_core_data_type import DataType
from ..core.plci_core_signal import Signal
from ..core.plci_core_scheduler_params import SchedulerParams

from .unit_test import get_boxes_libraries, get_schedulers_library, simulate

# Get library
boxesLibraries = get_boxes_libraries()
schedulersLibraries = get_schedulers_library()

# Scheduler parameters
scheduler : Scheduler = schedulersLibraries.construct_scheduler('Schedulers', 'Default')
scheduler.params = SchedulerParams(stop_time_ = 1.0, step_time_ = 0.001)

scheduler.diagram.append(Signal(0, DataType.FLOAT, 0.0))
scheduler.diagram.append(Signal(1, DataType.FLOAT, 0.0))
scheduler.diagram.append(Signal(2, DataType.FLOAT, 0.0))
scheduler.diagram.append(Signal(3, DataType.FLOAT, 0.0))
scheduler.diagram.append(Signal(4, DataType.FLOAT, 0.0))

# Sources
scheduler.diagram.append(boxesLibraries.construct_box('Boxes.Sources', 'Step', start_value = 0.0, stop_value = 1.0, step_time = 0.5))
scheduler.diagram.append(boxesLibraries.construct_box('Boxes.Sources', 'Ramp', start_time = 0.5, initial_value = 0.0, slope = 0.5))
scheduler.diagram.append(boxesLibraries.construct_box('Boxes.Sources', 'Constant', value = 7))

# Multiplier
scheduler.diagram.append(boxesLibraries.construct_box('Boxes.Math', 'Multiplier'))
scheduler.diagram.append(boxesLibraries.construct_box('Boxes.Math', 'Multiplier'))

# Sink
scheduler.diagram.append(boxesLibraries.construct_box('Boxes.Sinks', 'Graph', start_time = 0, stop_time = 1.0, title = "step box"))
scheduler.diagram.append(boxesLibraries.construct_box('Boxes.Sinks', 'Graph', start_time = 0, stop_time = 1.0, title = "ramp box"))
scheduler.diagram.append(boxesLibraries.construct_box('Boxes.Sinks', 'Graph', start_time = 0, stop_time = 1.0, title = "output"))
 
# Diagram creation
scheduler.diagram.add_box_outputs(scheduler.diagram.boxes[0], scheduler.diagram.signals[0])
scheduler.diagram.add_box_outputs(scheduler.diagram.boxes[1], scheduler.diagram.signals[1])
scheduler.diagram.add_box_outputs(scheduler.diagram.boxes[2], scheduler.diagram.signals[2])
scheduler.diagram.add_box_outputs(scheduler.diagram.boxes[3], scheduler.diagram.signals[3])
scheduler.diagram.add_box_outputs(scheduler.diagram.boxes[4], scheduler.diagram.signals[4])

scheduler.diagram.add_box_inputs(scheduler.diagram.boxes[3], scheduler.diagram.signals[0], scheduler.diagram.signals[1])
scheduler.diagram.add_box_inputs(scheduler.diagram.boxes[4], scheduler.diagram.signals[3], scheduler.diagram.signals[2])
scheduler.diagram.add_box_inputs(scheduler.diagram.boxes[5], scheduler.diagram.signals[0])
scheduler.diagram.add_box_inputs(scheduler.diagram.boxes[6], scheduler.diagram.signals[1])
scheduler.diagram.add_box_inputs(scheduler.diagram.boxes[7], scheduler.diagram.signals[4])


# Launch solver
simulate(scheduler)


#-----------------------------------
# Imports
#-----------------------------------

from typing import List

#-----------------------------------
# Functions
#-----------------------------------

def _init_step_box (box, simulation_params) -> None :

    if not 'step_time' in box :
        box['step_time'] = 1.0

    if not 'start_value' in box :
        box['start_value'] = 0.0

    if not 'stop_value' in box :
        box['stop_value'] = 1.0

def _function_step_box (box, event_) -> List : 
    if event_.timing < box['step_time']:
        v = box['start_value']
    else:
        v = box['stop_value']

    events : List = []
    
    for output in box.outputs :
        events.append(box.construct_signal_event(output, v))

    return events


#-----------------------------------
# Imports
#-----------------------------------

from typing import Callable, List, Any

from PyQt5.QtGui import QWindow
from muphyn.packages.core.box_library.plci_core_boxes_libraries import BoxesLibrariesManager
from muphyn.packages.core.scheduler_library.plci_core_schedulers_libraries import SchedulersLibraries
from muphyn.packages.core.utils.log_manager import LogManager

from muphyn.packages.interface.dialogs.library_dialog import LibraryDialog
from muphyn.packages.interface.dialogs.new_box_dialog import NewBoxDialog
from muphyn.packages.interface.dialogs.new_simulation_dialog import NewSimulationDialog
from muphyn.packages.interface.dialogs.simulation_parameters_dialog import SimulationParametersDialog
from muphyn.packages.interface.models.editable_models.simulation_model import SimulationModel
from muphyn.packages.interface.user_data.user_data import UserData

from .abstract_dialog import AbstractDialog

#-----------------------------------
# Class
# #-----------------------------------

class DialogsHolder :
    """Est la classe qui permet à une classe de fenêtre principale d'afficher des dialogues."""

    # -------------
    # Constructors
    # -------------

    def __init__ (self, main_window : QWindow) :
        self._dialogs : List[AbstractDialog] = []
        self._main_window : QWindow = main_window

        self._dialog_opened_listner : List[Callable[[Any], None]] = []
        self._dialog_closed_listner : List[Callable[[Any, Any], None]] = []
        self._dialog_closed_all_listener : List[Callable[[Any], None]] = []

    # -------------
    # Properties
    # -------------

    @property
    def dialogs (self) -> List[AbstractDialog] :
        """Permet de récuperer la liste des boîtes de dialogues actuellement affichées."""
        return self._dialogs

    @property
    def len (self) -> int :
        """Permet de récuperer le nombre de dialogues actuellement affichées."""
        return self._dialogs.__len__()

    @property
    def main_window (self) -> QWindow :
        """Permet de récuperer la fenêtre principale qui détient le dialog holder"""
        return self._main_window

    @property
    def dialog_opened_listener (self) -> List[Callable[[Any], None]] :
        """Permet de récuperer les écouteurs de l'événement appelés à l'ouverture d'une boite dialogue."""
        return self._dialog_opened_listner

    @property
    def dialog_closed_listener (self) -> List[Callable[[Any, Any], None]] :
        """Permet de récuperer les écouteurs de l'événement appelés à la fermeture d'une boite dialogue."""
        return self._dialog_closed_listner

    @property
    def dialog_closed_all_listener (self) -> List[Callable[[Any], None]] :
        """Permet de récuperer les écouteurs de l'événement appelés à l'ouverture d'une boite dialogue."""
        return self._dialog_closed_all_listener

    # -------------
    # Methods
    # -------------

    def _dialog_closed (self, dialog : AbstractDialog, answer : Any) -> None :
        """Est la méthode appelée par les boite de dialogue quand elle sont fermées."""
        
        if dialog in self._dialogs: 

            self._dialogs.remove(dialog)

            for dialog_listener in self._dialog_closed_listner :
                dialog_listener(dialog, answer)

    def __len__ (self) -> int :
        """Permet de récuperer le nombre de dialogues actuellement affichées."""
        return self._dialogs.__len__()

    def show_dialog (self, name : str, modal : bool, **kwargs) -> None : 
        """Permet d'afficher une nouvelle boite de dialogue."""
        
        dialog = None

        if name == 'library' :
            
            if 'boxes_libraries' in kwargs and 'schedulers_libraries' in kwargs :
            
                if isinstance(kwargs['boxes_libraries'], BoxesLibrariesManager) and isinstance(kwargs['schedulers_libraries'], SchedulersLibraries) :
                    dialog = LibraryDialog(self, kwargs['schedulers_libraries'])

        elif name == 'about' :

            LogManager().error('Not implemented : show about dialog')

        elif name == 'documentation' :

            LogManager().error('Not implemented : show documentation dialog')
            
        elif name == 'new_simulation' : 

            if 'schedulers_libraries' in kwargs :

                if isinstance(kwargs['schedulers_libraries'], SchedulersLibraries) :

                    dialog = NewSimulationDialog(self, kwargs['schedulers_libraries'])

        elif name == 'new_box' :

            LogManager().error('Not implemented : show new box dialog')
            """
            if 'user_data' in kwargs :

                if isinstance(kwargs['user_data'], UserData) :

                    dialog = NewBoxDialog(self, kwargs['user_data'])
            """

        elif name == 'new_scheduler' :

            LogManager().error('Not implemented : show new scheduler dialog')
            """if 'user_data' in kwargs :

                if isinstance(kwargs['user_data'], UserData) :
                    
                    dialog = NewSchedulerDialog(self, kwargs['user_data'])
            """

        elif name == 'simulation_parameters_dialog' : 

            if 'current_simulation' in kwargs and 'schedulers_libraries' in kwargs : 

                if isinstance(kwargs['current_simulation'], SimulationModel) and isinstance(kwargs['schedulers_libraries'], SchedulersLibraries):

                    dialog = SimulationParametersDialog(self, kwargs['current_simulation'], kwargs['schedulers_libraries'])

        if dialog is None: 
            return

        dialog.setModal(modal)

        self._dialogs.append(dialog)
        for dialog_listener in self._dialog_opened_listner :
            dialog_listener(dialog)

        if modal :
            self._dialog_closed(dialog, dialog.exec())
        else :
            dialog.show()

    def close_all (self) -> None :
        """Permet de fermer toutes les boite de dialogue."""

        for dialog in self._dialogs :
            dialog.close()

        for dialog_listener in self._dialog_closed_all_listener:
            dialog_listener(self)


#-----------------------------------
# Imports
#-----------------------------------

from typing import Any

from PyQt5 import QtCore, QtGui
from PyQt5.QtWidgets import QDialog

#-----------------------------------
# Class
#-----------------------------------

class AbstractDialog (QDialog) :
    """Est la classe abstraite commune aux dialogues affichés par dessus la fenêtre principale."""
    
    # -------------
    # Constructors
    # -------------

    def __init__ (self, dialog_holder : Any, name : str, title : str) :
        QDialog.__init__(self)

        self.setWindowTitle(title)
        self.setWindowFlag(QtCore.Qt.WindowContextHelpButtonHint, False)

        if not self.objectName():
            self.setObjectName('_dlg_' + name)
        
        self._name : str = name
        self._dialog_holder : Any = dialog_holder
        self._value : Any = None

    # -------------
    # Properties
    # -------------

    @property
    def name (self) -> str:
        """Permet de récuperer le nom de la boîte de dialogue."""
        return self._name

    @property
    def value (self) -> Any:
        """Permet de récuperer la valeur de la boite de dialogue."""
        return self._value

    # -------------
    # Methods
    # -------------

    def hideEvent (self, event: QtGui.QHideEvent) -> None:
        """Est la méthode appelée lorsque la boite de dialogue est fermée."""
        super().hideEvent(event)

        if not self.isModal():
            self._dialog_holder._dialog_closed(self, None)


import numpy as np
import sys
import re
from typing import Any, Union

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QIcon, QKeyEvent
from PyQt5.QtWidgets import QCheckBox, QDoubleSpinBox, QHBoxLayout, QLabel, QLineEdit, QSpinBox, QAction

from muphyn.packages.core.plci_core_data_type import DataType
from muphyn.packages.core.utils.log_manager import LogManager
from muphyn.packages.interface.properties_pages.data_property_widgets.abstract_property_widget import AbstractPropertyWidget
from muphyn.packages.interface.widgets.BaseWidgets.Button import ArrowButton
from muphyn.utils.paths import ROOT_DIR
from muphyn.packages.core.utils.global_env_variables_manager import GlobalEnvVariablesManager
from muphyn.packages.core.utils.regexes import *

class PropertyLineEdit(AbstractPropertyWidget):

    def __init__(self, parameter_to_edit, propertyType: DataType, parent = None, flags = Qt.Widget) -> None:
        super().__init__(parent, flags)

        # Init Layout
        layout = QHBoxLayout()

        # 
        self.parameter_to_edit = parameter_to_edit
        self.propertyType: DataType = propertyType

        # Init Line Edit
        self.line_edit = QLineEdit()
        self.line_edit.editingFinished.connect(self.line_edit_editing_finished)
        self.line_edit_tooltip = QAction(
            QIcon(ROOT_DIR + "/" +"packages/interface/assets/GeneralIcons/error.svg"), 
            ""
        )

        # Set Layout
        layout.addWidget(self.line_edit)
        self.setLayout(layout)

    def line_edit_editing_finished(self):
        self.setValue(self.line_edit.text())

    def setError(self, isError: bool, message: str = None):
        if isError:
            # Test if Tool Tip Icon is in Line Edit
            if not self.line_edit_tooltip in self.line_edit.actions():
                self.line_edit.addAction(
                    self.line_edit_tooltip, 
                    QLineEdit.ActionPosition.TrailingPosition
                )

            # Set Tool Tip message
            if message is not None:
                self.line_edit_tooltip.setToolTip(message)
            
        else:
            # Remove tooltip Icon
            if self.line_edit_tooltip in self.line_edit.actions():                
                self.line_edit.removeAction(self.line_edit_tooltip)
        
    def setValue(self, new_value: str):
        # Remove leading and ending whitespaces
        new_value = str(new_value).strip()

        # Handle Comma decimal separator
        if re.match(CommaScientificNumberRegex, new_value) or re.match(CommaFloatRegex, new_value):
            new_value = new_value.replace(",", ".")

        if self.propertyType == DataType.INT:
            # Test if empty string
            if new_value == "" or new_value == "+" or new_value == "-":
                # Valid Value
                self.validValue(0)

            # Test if floating value
            elif re.match(DotFloatRegex, new_value):
                # Valid Value
                self.validValue(int(float(new_value)))
            
            elif re.match(DotScientificNumberRegex, new_value):
                # Valid Value
                self.validValue(new_value)

            # Test if variable name exists 
            elif new_value in GlobalEnvVariablesManager().global_vars:
                # Get global variable
                global_var = GlobalEnvVariablesManager().global_vars[new_value]

                if type(global_var) == int or type(global_var) == float:
                    # Valid Value
                    self.validValue(new_value)
                else:
                    self.setError(True, f"Integer object expected: {type(global_var)} instead")
                    LogManager().error(TypeError(f"Integer object expected: {type(global_var)} instead"))
            else:
                self.setError(True, f"Invalid input: Please set a int value or an existing variable name")
                LogManager().error(TypeError(f"Invalid input: Not a int neither an existing variable name"))

        elif self.propertyType == DataType.FLOAT:

            # Test if empty string
            if new_value == "" or new_value == "+" or new_value == "-":
                # Valid Value
                self.validValue(0.0)

            # Test if point floating value
            elif re.match(DotFloatRegex, new_value):
                # Valid Value
                self.validValue(float(new_value))
            
            elif re.match(DotScientificNumberRegex, new_value):
                # Valid Value
                self.validValue(new_value)

            elif new_value in GlobalEnvVariablesManager().global_vars:
                # Get global variable
                global_var = GlobalEnvVariablesManager().global_vars[new_value]

                if type(global_var) == float or type(global_var) == int:
                    # Set Checked
                    self.validValue(new_value)
                else:
                    self.setError(True, f"Float object expected: {type(global_var)} instead")
                    LogManager().error(TypeError(f"Float object expected: {type(global_var)} instead"))
            else:
                self.setError(True, f"Invalid input: Please set a float value or an existing variable name")
                LogManager().error(TypeError(f"Invalid input: Not a float neither an existing variable name"))

    def validValue(self, new_value: Any):
        # Save value
        self._value = new_value

        # Set Checked
        self.line_edit.setText(str(new_value))

        # Emit value changed
        self.valueChanged.emit()

        # Reset Error
        self.setError(False)



class BooleanPropertyWidget(AbstractPropertyWidget):

    def __init__(self, parent = None, flags = Qt.Widget) -> None:
        super().__init__(parent, flags)
        
        # Value
        self._value = False

        # Init VBoxLayout
        layout = QHBoxLayout()

        # Double Spin
        self.check_box = QCheckBox()
        self.check_box.toggled.connect(self.on_check_box_value_changed)

        # Set Layout
        layout.addWidget(self.check_box)
        self.setLayout(layout)

    # -------------
    # Properties
    # -------------

    # -------------
    # Methods
    # -------------
    def on_check_box_value_changed(self, checked: bool):
        self.setValue(checked)

    def setValue(self, new_value: bool):
        if self._value != new_value:
            # Write new value
            self._value = new_value

            # Set Checked
            self.check_box.setCheckState(Qt.CheckState.Checked if new_value else Qt.CheckState.Unchecked)

            # Emit value changed
            self.valueChanged.emit()

class DoublePropertyWidget(AbstractPropertyWidget):

    # Constants
    DotFloatRegex = r"^[-+]?[0-9]*\.[0-9]*$"

    def __init__(self, parameter_to_edit: dict, parent = None, flags = Qt.Widget) -> None:
        super().__init__(parent, flags)

        # Value
        self._value = 0.0

        # Init VBoxLayout
        layout = QHBoxLayout()

        # Exponential value
        self.exp = 0

        # Get min/max values
        self.min_value = -sys.float_info.max
        if "min" in parameter_to_edit:
            min_value = parameter_to_edit["min"]
            try:
                self.min_value = max(float(min_value), self.min_value)
            except:
                LogManager().error(f"DoublePropertyWidget.__init__(): given 'min' value not a integer: {min_value}")


        self.max_value = sys.float_info.max
        if "max" in parameter_to_edit:
            max_value = parameter_to_edit["max"]
            try:
                self.max_value = min(float(max_value), self.max_value)
            except:
                LogManager().error(f"DoublePropertyWidget.__init__(): given 'max' value not a integer: {max_value}")


        # Double Spin
        self.double_spin_box = QDoubleSpinBox()
        self.double_spin_box.setRange(self.min_value, self.max_value)
        self.double_spin_box.valueChanged.connect(self.double_spin_box_value_changed)

        # Left arrow
        self.decrease_exp_button = ArrowButton(QIcon(ROOT_DIR + "/" + "packages/interface/assets/GeneralIcons/right_arrow.svg"), "")
        self.decrease_exp_button.clicked.connect(self.decrease_exp_button_clicked)

        # Right arrow
        self.increase_exp_button = ArrowButton(QIcon(ROOT_DIR + "/" + "packages/interface/assets/GeneralIcons/left_arrow.svg"), "")
        self.increase_exp_button.clicked.connect(self.increase_exp_button_clicked)

        # Text cursor event
        self.double_spin_box_line_edit = self.double_spin_box.lineEdit()
        self.double_spin_box_line_edit.editingFinished.connect(self.line_edit_editing_finished)
        self.double_spin_box_line_edit.returnPressed.connect(self.line_edit_editing_finished)

        # Set Layout
        layout.addWidget(self.double_spin_box, stretch=1)
        layout.addWidget(self.increase_exp_button)
        layout.addWidget(self.decrease_exp_button)
        self.setLayout(layout)

    # -------------
    # Properties
    # -------------

    # -------------
    # Methods
    # -------------
    def update_step_value(self):
        # Add point in text if doesn't exists
        double_spin_box_text = self.double_spin_box.cleanText()

        # If empty
        if double_spin_box_text == "":
            self.double_spin_box.setValue(0.0)
        # If integer
        elif re.match(IntegerRegex, double_spin_box_text):
            self.double_spin_box.setValue(float(self.double_spin_box.value()))

        # Get number of digits interger & decimal parts
        # If comma decimal separator
        if re.match(CommaFloatRegex, double_spin_box_text):
            integer_digits, decimal_digits = [len(part) for part in self.double_spin_box.cleanText().split(",", maxsplit=2)]
        
        # If dot decimal separator
        elif re.match(DotFloatRegex, double_spin_box_text):
            integer_digits, decimal_digits = [len(part) for part in self.double_spin_box.cleanText().split(".", maxsplit=2)]

        if self.exp >= 0:
            # Change value in integers
            if  self.exp >= integer_digits:
                # Block increasement of exp if exceed number of interger part digits
                self.exp = integer_digits - 1

            # Calculate cursor position
            cursor_position = integer_digits - self.exp - 1

            # Set Cursor
            self.double_spin_box_line_edit.setSelection(cursor_position, 1)
        else:
            # Change value in decimals

            # Absolute value of exponential
            abs_exp = abs(self.exp)
            if decimal_digits < abs_exp:
                # Update number of decimals in the DoubleSpinBox
                self.double_spin_box.setDecimals(abs_exp)

            # Calculate cursor position
            cursor_position = integer_digits + abs_exp

            # Set Cursor
            self.double_spin_box_line_edit.setSelection(cursor_position, 1)

        # Update DoubleSpinBox step size
        self.double_spin_box.setSingleStep(10**self.exp)

    def decrease_exp_button_clicked(self):
        self.exp -= 1
        self.update_step_value()

    def increase_exp_button_clicked(self):
        self.exp += 1
        self.update_step_value()

    def line_edit_editing_finished(self):
        # Set exp
        if self.exp > len(self.double_spin_box_line_edit.text()):
            self.setExp(len(self.double_spin_box_line_edit.text()))

    def double_spin_box_value_changed(self, new_value: float):
        self.setValue(new_value)

    def setExp(self, new_exp: int):
        if self.exp != new_exp:
            self.exp = new_exp
            self.update_step_value()

    def setValue(self, new_value: float):
        if self._value != new_value:
            # Write new value
            self._value = new_value

            # Set value on graphical element
            if new_value != self.double_spin_box.value():
                self.double_spin_box.setValue(new_value)

            # Emit value changed
            self.valueChanged.emit()

    def keyReleaseEvent(self, a0: QKeyEvent) -> None:
        if self.double_spin_box_line_edit.hasFocus():
            if a0.key() == Qt.Key.Key_Left:
                self.increase_exp_button_clicked()
            elif a0.key() == Qt.Key.Key_Right:
                self.decrease_exp_button_clicked()

        return super().keyReleaseEvent(a0)
            

class IntegerPropertyWidget(AbstractPropertyWidget):

    def __init__(self, parameter_to_edit, parent = None, flags = Qt.Widget) -> None:
        super().__init__(parent, flags)

        # Value
        self._value = 0.0

        # Init VBoxLayout
        layout = QHBoxLayout()

        # Exponential value
        self.exp = 0

        # Get min/max values
        int32_info = np.iinfo(np.int32)
        self.min_value = int32_info.min
        if "min" in parameter_to_edit:
            min_value = parameter_to_edit["min"]
            try:
                self.min_value = max(int(float(min_value)), self.min_value)
            except:
                LogManager().error(f"IntegerPropertyWidget.__init__(): given 'min' value not a integer: {min_value}")


        self.max_value = int32_info.max
        if "max" in parameter_to_edit:
            max_value = parameter_to_edit["max"]
            try:
                self.max_value = min(int(float(max_value)), self.max_value)
            except:
                LogManager().error(f"IntegerPropertyWidget.__init__(): given 'max' value not a integer: {max_value}")


        # Integer Spin
        self.integer_spin_box = QSpinBox()
        self.integer_spin_box.setRange(self.min_value, self.max_value)
        self.integer_spin_box.valueChanged.connect(self.integer_spin_box_value_changed)

        # Left arrow
        self.decrease_exp_button = ArrowButton(QIcon(ROOT_DIR + "/" + "packages/interface/assets/GeneralIcons/right_arrow.svg"), "")
        self.decrease_exp_button.clicked.connect(self.decrease_exp_button_clicked)

        # Right arrow
        self.increase_exp_button = ArrowButton(QIcon(ROOT_DIR + "/" + "packages/interface/assets/GeneralIcons/left_arrow.svg"), "")
        self.increase_exp_button.clicked.connect(self.increase_exp_button_clicked)

        # Text cursor event
        self.integer_spin_box_line_edit = self.integer_spin_box.lineEdit()
        self.integer_spin_box_line_edit.editingFinished.connect(self.line_edit_editing_finished)
        self.integer_spin_box_line_edit.returnPressed.connect(self.line_edit_editing_finished)

        # Set Layout
        layout.addWidget(self.integer_spin_box, stretch=1)
        layout.addWidget(self.increase_exp_button)
        layout.addWidget(self.decrease_exp_button)
        self.setLayout(layout)

    # -------------
    # Properties
    # -------------

    # -------------
    # Methods
    # -------------
    def update_step_value(self):
        # Get number of digits interger & decimal parts
        integer_digits = len(self.integer_spin_box.cleanText())

        # Change value in integers
        if  self.exp >= integer_digits:
            # Block increasement of exp if exceed number of interger part digits
            self.exp = integer_digits - 1

        # Calculate cursor position
        cursor_position = integer_digits - self.exp - 1

        # Set Cursor
        self.integer_spin_box_line_edit.setSelection(cursor_position, 1)

        # Update DoubleSpinBox step size
        self.integer_spin_box.setSingleStep(10**self.exp)

    def decrease_exp_button_clicked(self):
        self.setExp(self.exp - 1)

    def increase_exp_button_clicked(self):
        self.setExp(self.exp + 1)

    def integer_spin_box_value_changed(self, new_value: float):
        self.setValue(new_value)

    def line_edit_editing_finished(self):
        if self.exp > len(self.integer_spin_box_line_edit.text()):
            self.setExp(len(self.integer_spin_box_line_edit.text()))

    def setExp(self, new_exp: int):
        if new_exp < 0:
            new_exp = 0

        if self.exp != new_exp:
            self.exp = new_exp
            self.update_step_value()

    def setValue(self, new_value: Union[str, int]):
        if new_value == "":
            new_value = 0

        if type(new_value) == str:
            new_value = int(new_value)

        if self._value != new_value:
            # Write new value
            self._value = new_value

            # Set value on graphical element
            if new_value != self.integer_spin_box.value():
                self.integer_spin_box.setValue(new_value)

            # Emit value changed
            self.valueChanged.emit()

    def keyReleaseEvent(self, a0: QKeyEvent) -> None:
        if self.integer_spin_box_line_edit.hasFocus():
            if a0.key() == Qt.Key.Key_Left:
                self.increase_exp_button_clicked()
            elif a0.key() == Qt.Key.Key_Right:
                self.decrease_exp_button_clicked()

        return super().keyReleaseEvent(a0)
            
class StringPropertyWidget(AbstractPropertyWidget):

    def __init__(self, parameter_to_edit: dict, parent = None, flags = Qt.Widget) -> None:
        super().__init__(parent, flags)
        
        # Value
        self._value = False

        # Init VBoxLayout
        layout = QHBoxLayout()

        # Set limit characters
        self.limit_characters = -1
        if "maxLength" in parameter_to_edit:
            limit_characters = parameter_to_edit["maxLength"]
            try:
                self.limit_characters = max(int(limit_characters), self.limit_characters)
            except:
                LogManager().error(f"StringPropertyWidget.__init__(): given 'maxLength' value not a valid value: {limit_characters}")

        # Double Spin
        self.line_edit = QLineEdit()
        self.line_edit.setMaxLength(self.limit_characters)
        self.line_edit.editingFinished.connect(self.line_edit_editing_finished)

        # Set Layout
        layout.addWidget(self.line_edit)
        self.setLayout(layout)

    # -------------
    # Properties
    # -------------

    # -------------
    # Methods
    # -------------
    def line_edit_editing_finished(self):
        self.setValue(self.line_edit.text())

    def setValue(self, new_value: str):
        if self._value != new_value:
            # Write new value
            self._value = new_value

            # Set Checked
            self.line_edit.setText(new_value)

            # Emit value changed
            self.valueChanged.emit()

class UnknownTypePropertyWidget(AbstractPropertyWidget):

    def __init__(self, type_name: str, parent = None, flags = Qt.Widget) -> None:
        super().__init__(parent, flags)
        
        # Value
        self._value = False

        # Type Name
        self._type_name = type_name

        # Init VBoxLayout
        layout = QHBoxLayout()

        # Double Spin
        self.label = QLabel(f"Uknown Type : {type_name}")

        # Set Layout
        layout.addWidget(self.label)
        self.setLayout(layout)

    # -------------
    # Properties
    # -------------

    # -------------
    # Methods
    # -------------
    def setValue(self, new_value: str):
        LogManager().error(f"Can't set value of UnknownType : {self._type_name}")


#-----------------------------------
# Imports
#-----------------------------------

#-----------------------------------
# Class
#-----------------------------------

from muphyn.packages.interface.models.editable_models.abstract_editable_model import AbstractEditableModel


class AbstractExporter :
    """Est la classe abstraite commune aux classes capables d'exporter des modèles."""

    # -------------
    # Constructors
    # -------------

    def __init__ (self) :
        ...

    # -------------
    # Methods
    # -------------

    def save (self, model : AbstractEditableModel, path : str) -> bool :
        """Est la méthode appelée pour sauvegarder le modèle éditable."""
        raise Exception("AbstractExporter save is an abstract method and must be overloaded.")



from ..core.plci_core_data_type import DataType
from ..core.plci_core_signal import Signal
from ..core.plci_core_box import Box
from ..core.plci_core_diagram import Diagram

s0 = Signal(0, DataType.FLOAT, 0)
s1 = Signal(1, DataType.FLOAT, 0)

b0 = Box(0, 'a', 'e')
b1 = Box(1, 'b', 'e')
b2 = Box(2, 'c', 'e')

d = Diagram()
d.append(s0)
d.append(s1)
d.append(b0)
d.append(b1)
d.append(b2)

d.add_box_outputs(d.boxes[0], d.signals[0])
d.add_box_outputs(d.boxes[1], d.signals[1])

d.add_box_inputs(d.boxes[1], d.signals[0])
d.add_box_inputs(d.boxes[2], d.signals[1])

print("Boxes : ")
for box in d._boxes:
    print("\t - ", box)

print("Signals : ")
for signal in d._signals:
    print("\t - ", signal)

for signal in d._box_inputs:
    print("signal " + str(signal.index) + " : ")

    for box in d._box_inputs[signal] :
        print("\t - Box " + str(box.index))



#-----------------------------------
# Imports
#-----------------------------------

# General Imports

# Project Imports
from PyQt5.QtCore import QSizeF
from PyQt5.QtWidgets import QWidget

# Project Imports
from muphyn.packages.interface.models.graphical_models.abstract_moveable_graphical_element import AbstractMoveableGraphicalElement
from muphyn.packages.interface.models.links_model.abstract_link_model import AbstractLinkModel
from muphyn.packages.interface.models.links_model.link_type import LinkType
from muphyn.packages.interface.models.signals_model.abstract_connection_model import AbstractConnectionModel

#-----------------------------------
# Class
#-----------------------------------
class TemporaryLink(AbstractLinkModel):
    def __init__(self, mouse_item: QWidget, signal_creator: AbstractConnectionModel):

        # Init Input Model (end of connector)
        input_model = AbstractMoveableGraphicalElement('mouse', mouse_item.pos(), QSizeF(0, 0))
        mouse_item.changeEvent()

        super().__init__(input_model, signal_creator, LinkType.SQUARE)

        



#-----------------------------------
# Import
#-----------------------------------

from enum import Enum
from typing import Any
    
#-----------------------------------
# Class
#-----------------------------------

class DataType(Enum) : 
    """Est l'enum qui décrit les types acceptés dans les signaux."""


    UNDIFINED = -1
    """Est le type qui définis un objet indéfinis."""

    STRING = 0,
    """Est le type qui définis un string."""

    INT = 1,
    """Est le type qui définis un entier."""

    FLOAT = 2,
    """Est le type qui définis un nombre à virgule flottante."""

    BOOLEAN = 3,
    """Est le type qui définis un boolean."""
    
    OBJECT = 4,
    """Est le type qui définis un objet."""

    ANYFILE = 5
    """Est le type qui définis un chemin vers un fichier existant ou non."""

    DIRECTORY = 6
    """Est le type qui définis un chemin vers un dossier existant."""

    EXISTINGFILE = 7
    """Est le type qui définis un chemin vers un fichier existant."""

    EXISTINGFILES = 8
    """Est le type qui définis une liste de chemins pointant chacun vers un fichier existant."""

    CHOICE = 9

    def __str__ (self) :
        """Permet de retourner le data type sous la forme d'un nom."""
        return self.name.lower()

    def default_value (self) -> Any :
        """Permet de récuperer la valeur par défaut des données suivant le type."""

        if self == DataType.STRING :
            return ""

        elif self == DataType.INT :
            return 0

        elif self == DataType.FLOAT :
            return 0.0

        elif self == DataType.BOOLEAN :
            return False

        elif self == DataType.ANYFILE :
            return ""

        elif self == DataType.DIRECTORY :
            return ""

        elif self == DataType.EXISTINGFILE :
            return ""

        elif self == DataType.EXISTINGFILES :
            return []

        elif self == DataType.CHOICE:
            return []
        
        else:
            return None


#-----------------------------------
# Functions
#-----------------------------------

def get_data_type (value) -> DataType :
    """Permet de retourner un data type suivant la valeur passé en paramètre."""
    if isinstance(value, str) :
        
        for type in DataType :
            if type.__str__() == value.lower() :

                return type

    elif isinstance(value, int) :
        return DataType[value]

    elif isinstance(value, DataType) : 
        return value
    
    return DataType.UNDIFINED


import sys
from PyQt5 import Qt
from PyQt5.QtWidgets import QApplication, QWidget, QPushButton, QLineEdit, QLabel, \
                            QHBoxLayout, QVBoxLayout


class MyApp(QWidget):
    def __init__(self):
        super().__init__()
        # self.window_width, self.window_height = 1200, 800
        # self.setMinimumSize(self.window_width, self.window_height)
        self.setMinimumWidth(1200)
        self.setMaximumHeight(200)
        self.setWindowTitle('Copy to Clipboard Example')
        self.setStyleSheet('height: 65px; font-size: 35px')

        self.layout = QVBoxLayout()
        self.setLayout(self.layout)

        self.lineEdit = QLineEdit()
        self.btn_copy = QPushButton('Copy to clipboard', clicked=self.copyToClipboard)
        self.btn_paste = QPushButton('Paste from clipboard', clicked=self.pasteFromClipboard)
        self.lbl = QLabel()

        self.layout.addWidget(self.lineEdit)
        self.layout.addWidget(self.btn_copy)
        self.layout.addWidget(self.btn_paste)
        self.layout.addWidget(self.lbl)

    def copyToClipboard(self):
        cb = QApplication.clipboard()
        cb.clear(mode=cb.Clipboard)
        cb.setText(self.lineEdit.text(), mode=cb.Clipboard)
        self.lbl.setText('Content is copied')

    def pasteFromClipboard (self) :
        cb = QApplication.clipboard()
        txt = cb.text(mode=cb.Clipboard)
        print('clipboard :', txt)

if __name__ == '__main__':
    # don't auto scale when drag app to a different monitor.
    
    app = QApplication(sys.argv)
    app.setAttribute(Qt.HighDpiScaleFactorRoundingPolicy.PassThrough)

    myApp = MyApp()
    myApp.show()

    try:
        sys.exit(app.exec_())
    except SystemExit:
        print('Closing Window...')


#-----------------------------------
# Imports
#-----------------------------------

from ..core.plci_core_signal_event import SignalEvent
from ..core.plci_core_scheduler import Scheduler
from ..core.plci_core_data_type import DataType
from ..core.plci_core_data import Data
from ..core.plci_core_signal import Signal
from ..core.plci_core_scheduler_params import SchedulerParams
from ..core.plci_core_diagram import Diagram

from .unit_test import get_boxes_libraries, get_schedulers_library

# Get library
boxesLibraries = get_boxes_libraries()
schedulersLibraries = get_schedulers_library()

d = Diagram()
d.append(Signal(index_ = 0, signal_type_ = DataType.FLOAT, default_value_ = 0.0))
d.append(Signal(index_ = 1, signal_type_ = DataType.FLOAT, default_value_ = 0.0))
d.append(Signal(index_ = 2, signal_type_ = DataType.FLOAT, default_value_ = 0.0))

d.append(boxesLibraries.construct_box('Boxes.Math', 'Multiplier'))
d.append(boxesLibraries.construct_box('Boxes.Sources', 'Constant', value = 0.0))
 
# Diagram creation

d.add_box_inputs(d.boxes[0], d.signals[0], d.signals[1])
d.add_box_outputs(d.boxes[0], d.signals[2])

scheduler : Scheduler = schedulersLibraries.construct_scheduler('Schedulers', 'Default')
scheduler.diagram = d
scheduler.params = SchedulerParams(10.0, 0.001)

# test output
value_0 : float = 10.0
value_1 : float = -5.0
answer : float = value_0 - value_1

se0 = scheduler.construct_event(None)
return_data = d.boxes[0].function(se0)

print("box current timing : " + str(d.boxes[0].current_timing))
print("box wait for events : " + str(d.boxes[0].wait_for_events))
print("box wait for all signal events : " + str(d.boxes[0].wait_for_all_signal_events))
print("box events on all inputs : " + str(d.boxes[0].events_on_all_inputs))
print("retrun data : " + str(return_data))
if len(return_data) == 0 :
    print("return data : no data")
else:
    print("return data new value : " + str(return_data[0].new_signal_value))

print("========================================================")

se1 = scheduler.construct_event(SignalEvent(d.signals[0], d.boxes[1], Data(DataType.FLOAT, value_0)))
se1.signal.data = se1.new_signal_data
return_data = d.boxes[0].function(se1)

print("box current timing : " + str(d.boxes[0].current_timing))
print("box wait for events : " + str(d.boxes[0].wait_for_events))
print("box wait for all signal events : " + str(d.boxes[0].wait_for_all_signal_events))
print("box events on all inputs : " + str(d.boxes[0].events_on_all_inputs))
print("retrun data : " + str(return_data))
if len(return_data) == 0 :
    print("return data : no data")
else:
    print("return data new value : " + str(return_data[0].new_signal_value))

print("========================================================")

scheduler.stepping_time()

se2 = scheduler.construct_event(SignalEvent(d.signals[0], d.boxes[1], Data(DataType.FLOAT, value_0)))
se2.signal.data = se2.new_signal_data
return_data = d.boxes[0].function(se2)

print("box current timing : " + str(d.boxes[0].current_timing))
print("box wait for events : " + str(d.boxes[0].wait_for_events))
print("box wait for all signal events : " + str(d.boxes[0].wait_for_all_signal_events))
print("box events on all inputs : " + str(d.boxes[0].events_on_all_inputs))
print("retrun data : " + str(return_data))
if len(return_data) == 0 :
    print("return data : no data")
else:
    print("return data new value : " + str(return_data[0].new_signal_value))

print("========================================================")

se3 = scheduler.construct_event(SignalEvent(d.signals[1], d.boxes[1], Data(DataType.FLOAT, value_1)))
se3.signal.data = se3.new_signal_data
return_data = d.boxes[0].function(se3)

print("box current timing : " + str(d.boxes[0].current_timing))
print("box wait for events : " + str(d.boxes[0].wait_for_events))
print("box wait for all signal events : " + str(d.boxes[0].wait_for_all_signal_events))
print("box events on all inputs : " + str(d.boxes[0].events_on_all_inputs))
print("retrun data : " + str(return_data))
if len(return_data) == 0 :
    print("return data : no data")
else:
    print("return data new value : " + str(return_data[0].new_signal_value))

print("========================================================")


#-----------------------------------
# Imports
#-----------------------------------

from typing import Iterable, List, Dict
from os import listdir, path

from muphyn.packages.core.plci_core_scheduler import Scheduler

from muphyn.packages.core.scheduler_library.scheduler_library_data import SchedulerData
from muphyn.packages.core.scheduler_library.abstract_scheduler_library_importer import AbstractSchedulerLibraryImporter
from muphyn.packages.core.scheduler_library.scheduler_library_importer_1_0_0 import SchedulerLibraryImporter_1_0_0
from muphyn.packages.core.utils.log_manager import LogManager

#-----------------------------------
# Classes
#-----------------------------------

class SchedulerLibraryElement : 
    """Est la classe qui permet de stocker les éléments de bibliothèque pour les planificateurs."""

    # -------------
    # Constructors
    # -------------
    def __init__ (self, path : str) :
        self._path = path
        self._loaded = False
        self._schedulers : Dict[str, SchedulerData] = {}

    # -------------
    # Properties
    # -------------
    
    @property
    def path (self) -> str : 
        """Permet de récuperer le lien vers le dossier de la bibliothèque de planificateurs.""" 
        return self._path

    @property
    def loaded (self) -> bool :
        """Permet de récuperer l'état de chargement de la bibliothèque de planificateurs."""
        return self._loaded

    @property
    def schedulers (self) -> Dict[str, SchedulerData] :
        """Permet de récuperer le dictionnaire des nom de bibliothèque et leur éléments de création de planificateurs."""
        return self._schedulers

    # -------------
    # Methods
    # -------------

    def load (self, scheduler_importer : AbstractSchedulerLibraryImporter) :
        """Permet de charger la bibliothèque."""
        for f in listdir(self.path) :

            if f.endswith('.yaml') :

                file_name = f.split('.')[0]
                absolute_yaml_file = self.path + '/' + file_name + '.yaml'

                imported_scheduler = scheduler_importer.import_scheduler_data(self.path, file_name, absolute_yaml_file, self._schedulers)

                if imported_scheduler is None :
                    continue

                if imported_scheduler['scheduler_data'] is None :
                    continue
                    
                self._schedulers[imported_scheduler['library_name']] = imported_scheduler['scheduler_data']

                        

class SchedulersLibraries :
    """Est la classe qui permet de construire les planificateurs.""" 

    # -------------
    # Constructors
    # -------------
    def __init__ (self) :
        self._libraries : List[SchedulerLibraryElement] = []
        self._scheduler_importer : AbstractSchedulerLibraryImporter = SchedulerLibraryImporter_1_0_0()
    
    # -------------
    # Properties
    # -------------

    @property
    def libraries (self) -> List[SchedulerLibraryElement] :
        """Permet de retourner la liste des libraries."""
        return self._libraries

    @property
    def schedulers (self) -> Iterable[SchedulerData] :
        """Permet de retourner la liste des planificateur de toutes les bibliothèques."""
        for schedulers_library in self.libraries :
            for scheduler_name in schedulers_library.schedulers :
                yield schedulers_library.schedulers[scheduler_name]
        
    @property 
    def scheduler_importer (self) -> AbstractSchedulerLibraryImporter :
        """Permet de retourner l'importeur de planificateurs.""" 
        return self._scheduler_importer

    # -------------
    # Methods
    # -------------
    def add_library (self, library_folder : str) :
        """Permet d'ajouter une bibliothèque dans le répertoire des bibliothèque."""

        # Test if library_folder path format is string 
        if not (isinstance(library_folder, str)) :
            LogManager().error(f"Wrong path variable format {type(library_folder)} instead of str")
            return False

        # Test if library folder path exists
        if not path.exists(library_folder):
            LogManager().error(f"Path doesn't exist : {library_folder}")
            return False

        # Check if library has already been added
        if any(libraryElement.path == library_folder for libraryElement in self._libraries):
            LogManager().info(f"Library has already been added : {library_folder}")
            return False

        # Append Scheduler Library
        self._libraries.append(SchedulerLibraryElement(library_folder))

        return True

    def load_libraries (self) :
        """Permet de charger toutes les bibliothèques du répertoire."""

        for libraryElement in self._libraries :

            if not libraryElement.loaded :
                libraryElement.load(self.scheduler_importer)

    def construct_scheduler (self, scheduler_library : str, scheduler_name : str) -> Scheduler:
        """Permet de construire un scheduler suivant son nom et sa librairie."""

        if not (isinstance(scheduler_library, str) and isinstance(scheduler_name, str)) :
            return None

        scheduler_access = self._name_library(scheduler_library, scheduler_name)

        for libraryElement in self._libraries :
            
            if scheduler_access in libraryElement.schedulers :
                return libraryElement.schedulers[scheduler_access].construct_scheduler()

    def _name_library (self, scheduler_library : str, schdeuler_name : str) -> str :
        """Permet de rassembler le nom et la libraire en un seul string."""
        return scheduler_library + "." + schdeuler_name

    def clear (self) -> None :
        """Permet d'éffacer le contenu des solveurs chargées."""
        for libraryElement in self._libraries :
            del libraryElement
        
        del self._libraries
        self._libraries : List[SchedulerLibraryElement] = []


import sys
import typing

from PyQt5 import QtGui
from PyQt5 import QtCore
from PyQt5.QtCore import QPointF, QRectF, QSizeF, Qt, pyqtSignal, pyqtSlot
from PyQt5.QtWidgets import QApplication, QGraphicsItem, QGraphicsObject, QGraphicsRectItem, QGraphicsScene, QGraphicsSceneMouseEvent, QGraphicsView
from muphyn.packages.interface.models.graphical_models.resizers.abstract_resizer import AbstractResizer

from muphyn.packages.interface.models.event_signal_data import EventSignalData
from muphyn.packages.interface.models.graphical_models.resizers.horizontal_resizer import HorizontalResizer
from muphyn.packages.interface.models.graphical_models.resizers.oblique_resizer import ObliqueResizer
from muphyn.packages.interface.models.graphical_models.resizers.vertical_resizer import VerticalResizer

from muphyn.packages.interface.models.graphical_models.box_model import BoxModel

if __name__ == "__main__":

    app = QApplication(sys.argv)

    view = QGraphicsView()
    scene = QGraphicsScene()
    scene.setSceneRect(0, 0, 1024, 768)
    view.setScene(scene)

    margin = (1024 - 400) / 5
    box000 = BoxModel('BoxLibrary', 'box 000°', QPointF((1 * margin) + 000, -50 + 384), QSizeF(100, 100), 0)
    box090 = BoxModel('BoxLibrary', 'box 090°', QPointF((2 * margin) + 100, -50 + 384), QSizeF(100, 100), 90)
    box180 = BoxModel('BoxLibrary', 'box 180°', QPointF((3 * margin) + 200, -50 + 384), QSizeF(100, 100), 180)
    box270 = BoxModel('BoxLibrary', 'box 270°', QPointF((4 * margin) + 300, -50 + 384), QSizeF(100, 100), 270)

    scene.addItem(box000)
    scene.addItem(box090)
    scene.addItem(box180)
    scene.addItem(box270)

    view.show()

    sys.exit(app.exec_())


# SOURCES : https://stackoverflow.com/questions/53349623/how-to-change-languagestranslations-dynamically-on-pyqt5

import sys
from PyQt5 import QtCore, QtGui, QtWidgets

class Demo (QtWidgets.QWidget):
    def __init__(self):
        super(Demo, self).__init__()
        self.button = QtWidgets.QPushButton()
        self.label = QtWidgets.QLabel(alignment=QtCore.Qt.AlignCenter)

        self.combo = QtWidgets.QComboBox(self)
        self.combo.currentIndexChanged.connect(self.change_func)

        self.trans = QtCore.QTranslator(self)

        self.v_layout = QtWidgets.QVBoxLayout(self)
        self.v_layout.addWidget(self.combo)
        self.v_layout.addWidget(self.button)
        self.v_layout.addWidget(self.label)

        options = ([('English', ''), ('français', 'eng-fr' ), ('中文', 'eng-chs'), ])
        
        for i, (text, lang) in enumerate(options):
            self.combo.addItem(text)
            self.combo.setItemData(i, lang)
        self.retranslateUi()

    @QtCore.pyqtSlot(int)
    def change_func(self, index):
        data = self.combo.itemData(index)
        if data:
            self.trans.load(data)
            QtWidgets.QApplication.instance().installTranslator(self.trans)
        else:
            QtWidgets.QApplication.instance().removeTranslator(self.trans)

    def changeEvent(self, event):
        if event.type() == QtCore.QEvent.LanguageChange:
            self.retranslateUi()
        super(Demo, self).changeEvent(event)

    def retranslateUi(self):
        self.button.setText(QtWidgets.QApplication.translate('Demo', 'Start'))
        self.label.setText(QtWidgets.QApplication.translate('Demo', 'Hello, World'))

if __name__ == '__main__':
    app = QtWidgets.QApplication(sys.argv)
    demo = Demo()
    demo.show()
    sys.exit(app.exec_())


#-----------------------------------
# Import
#-----------------------------------

from enum import Enum
    
#-----------------------------------
# Class
#-----------------------------------

class LinkType(Enum) : 
    """Est l'énumeration qui reprends les différents mode de méthode de dessin des signaux."""

    SQUARE = 0,
    """Est le mode de lien où les signaux doivent décrire des angles droits."""

    CURVED = 1
    """Est le mode de lien où les signaux doivent décrire des courbes."""

    def __str__ (self) -> str :

        if self == LinkType.SQUARE :
            return 'SQUARE'

        else :
            return 'CURVED'
    
    
#-----------------------------------
# Methods
#-----------------------------------

def get_link_type (link_type : str) -> LinkType :

    if link_type == 'SQUARE' :
        return LinkType.SQUARE

    else :
        return LinkType.CURVED


#-----------------------------------
# Imports
#-----------------------------------

import os
from datetime import date
from typing import Any

from PyQt5.QtCore import QCoreApplication, QRect, pyqtSlot
from PyQt5.QtWidgets import QComboBox, QFileDialog, QFrame, QLabel, QLineEdit, QGridLayout, QVBoxLayout, QPushButton, QWidget
from muphyn.packages.core.plci_core_scheduler_params import SchedulerParams

from muphyn.packages.core.scheduler_library.plci_core_schedulers_libraries import SchedulersLibraries
from muphyn.packages.core.scheduler_library.scheduler_library_data import SchedulerData
from muphyn.packages.interface.models.dbl_spin_box import DblSpinBox
from muphyn.packages.interface.models.editable_models.scheduler_model import SchedulerModel
from muphyn.packages.interface.models.editable_models.simulation_model import SimulationModel
from .abstract_dialog import AbstractDialog
from muphyn.packages.interface.widgets.BaseWidgets.Button import PlainButton
from muphyn.packages.interface.widgets.BaseWidgets.Label import PropertyLabel
from muphyn.packages.interface.utils.interface_constants import ApplicationWindowTitle

#-----------------------------------
# Class
#-----------------------------------

class NewSimulationDialog (AbstractDialog) :
    """Est la classe permettant d'afficher une boîte de dialogue capable de créer une simulation."""
    
    # -------------
    # Constructors
    # -------------

    def __init__ (self, dialog_holder : Any, schedulers_libraries : SchedulersLibraries) :
        AbstractDialog.__init__(self, dialog_holder, 'new_simulation', 'Nouvelle Simulation')

        self.setFixedSize(480, 240)
        self._init_ui()
        self._test_data_accept_button()
        
        self._schedulers_libraries = schedulers_libraries

        for scheduler_data in self._schedulers_libraries.schedulers :
            self._cmb_scheduler.addItem(scheduler_data.__str__(), scheduler_data)


    # -------------
    # Methods
    # -------------

    @pyqtSlot()
    def _btn_cancel_click (self) -> None :
        """Est la méthode appelée lorsque l'utilisateur clique sur le bouton annuler."""
        self.close()
        
    @pyqtSlot()
    def _btn_simulation_path_click (self) -> None :
        """Est la méthode appelée lorsque l'utilisateur clique sur le bouton rechercher."""
        path = QFileDialog.getExistingDirectory(self, f"{ApplicationWindowTitle} - Select folder", os.getcwd())

        if path is None or path == '':
            return

        if not path.endswith('/') :
            path = path + '/'

        self._fld_simulation_path.setText(path)

    @pyqtSlot()
    def _test_data_accept_button (self) -> None :
        """Permet de vérifier si les données entrées sont correctes dans le but de débloquer le bouton accepter."""
        
        if self._fld_simulation_name.text().strip().__len__() == 0 :
            self._btn_accept.setEnabled(False)
            return

        if not os.path.isdir(self._fld_simulation_path.text()) :
            self._btn_accept.setEnabled(False)
            return

        if self._cmb_scheduler.currentIndex() == -1 :
            self._btn_accept.setEnabled(False)
            return
        
        if self._spn_simulation_timing.value() == 0 :
            self._btn_accept.setEnabled(False)
            return

        if self._spn_simulation_step.value() == 0 :
            self._btn_accept.setEnabled(False)
            return
            
        self._btn_accept.setEnabled(True)


    def _init_ui (self) :

        # Base Layout
        baseLayout = QVBoxLayout()

        # Options Layout
        optionsLayout = QGridLayout()
        optionsLayout.setColumnMinimumWidth(2, 90)
        optionsLayout.setSpacing(5)

        # Buttons Layout
        buttonsLayout = QGridLayout()
        buttonsLayout.setColumnMinimumWidth(1, 90)
        buttonsLayout.setColumnMinimumWidth(2, 90)
        buttonsLayout.setHorizontalSpacing(5)

        # Simulation Name
        self._lbl_simulation_name = PropertyLabel()
        self._lbl_simulation_name.setObjectName("_lbl_simulation_name")

        self._fld_simulation_name = QLineEdit()
        self._fld_simulation_name.setObjectName("_fld_simulation_name")
        self._fld_simulation_name.textChanged.connect(self._test_data_accept_button)

        # Simulation Path
        self._lbl_simulation_path = PropertyLabel()
        self._lbl_simulation_path.setObjectName("_lbl_simulation_path")
        
        self._fld_simulation_path = QLineEdit()
        self._fld_simulation_path.setObjectName("_fld_simulation_path")
        self._fld_simulation_path.textChanged.connect(self._test_data_accept_button)

        self._btn_simulation_path = QPushButton()
        self._btn_simulation_path.setObjectName("_btn_simulation_path")
        self._btn_simulation_path.clicked.connect(self._btn_simulation_path_click)

        # Separator
        self._ln_separator = QFrame()
        self._ln_separator.setObjectName("_ln_separator")
        self._ln_separator.setFrameShape(QFrame.HLine)
        self._ln_separator.setFrameShadow(QFrame.Sunken)

        # Scheduler
        self._lbl_scheduler = PropertyLabel()
        self._lbl_scheduler.setObjectName("_lbl_scheduler")

        self._cmb_scheduler = QComboBox()
        self._cmb_scheduler.setObjectName("_cmb_scheduler")
        self._cmb_scheduler.currentIndexChanged.connect(self._test_data_accept_button)

        # Simulation Time
        self._lbl_simulation_timing = PropertyLabel()
        self._lbl_simulation_timing.setObjectName("_lbl_simulation_timing")

        self._spn_simulation_timing = DblSpinBox()
        self._spn_simulation_timing.setObjectName("_spn_simulation_timing")
        self._spn_simulation_timing.setMinimum(0)
        self._spn_simulation_timing.setValue(5.0)
        self._spn_simulation_timing.valueChanged.connect(self._test_data_accept_button)

        # Step Time
        self._lbl_simulation_step = PropertyLabel()
        self._lbl_simulation_step.setObjectName("_lbl_simulation_step")

        self._spn_simulation_step = DblSpinBox()
        self._spn_simulation_step.setObjectName("_spn_simulation_step")
        self._spn_simulation_step.setMinimum(0)
        self._spn_simulation_step.setValue(0.001)
        self._spn_simulation_step.setSingleStep(self._spn_simulation_step.value())
        self._spn_simulation_step.valueChanged.connect(self._test_data_accept_button)

        # Confirm Button
        self._btn_accept = PlainButton()
        self._btn_accept.setObjectName("_btn_accept")
        self._btn_accept.clicked.connect(self._btn_accept_click)

        # Cancel Button
        self._btn_cancel = QPushButton()
        self._btn_cancel.setObjectName("_btn_cancel")
        self._btn_cancel.clicked.connect(self._btn_cancel_click)

        # Add option widgets to layout
        optionsLayout.addWidget(self._lbl_simulation_name, 0, 0)
        optionsLayout.addWidget(self._fld_simulation_name, 0, 1)
        optionsLayout.addWidget(self._lbl_simulation_path, 1, 0)
        optionsLayout.addWidget(self._fld_simulation_path, 1, 1)
        optionsLayout.addWidget(self._btn_simulation_path, 1, 2)
        optionsLayout.addWidget(self._ln_separator, 2, 0, 1, 3)
        optionsLayout.addWidget(self._lbl_scheduler, 3, 0)
        optionsLayout.addWidget(self._cmb_scheduler, 3, 1)
        optionsLayout.addWidget(self._lbl_simulation_timing, 4, 0)
        optionsLayout.addWidget(self._spn_simulation_timing, 4, 1)
        optionsLayout.addWidget(self._lbl_simulation_step, 5, 0)
        optionsLayout.addWidget(self._spn_simulation_step, 5, 1)

        # Add buttons to layout
        buttonsLayout.addWidget(self._btn_accept, 0, 1)
        buttonsLayout.addWidget(self._btn_cancel, 0, 2)
        buttonsLayout.setColumnStretch(0, 1)

        # Add layouts to main layout
        baseLayout.addLayout(optionsLayout)
        baseLayout.addLayout(buttonsLayout)
        
        # Set Base Layout of page
        self.setLayout(baseLayout)

        self.retranslateUi(self)

    def retranslateUi (self, Dialog) -> None :
        self.setWindowTitle(QCoreApplication.translate(self.objectName(), f"{ApplicationWindowTitle} - New Simulation", None))
        self._btn_accept.setText(QCoreApplication.translate("_dlg_new_simulation", "Accept", None))
        self._btn_cancel.setText(QCoreApplication.translate("_dlg_new_simulation", "Cancel", None))
        self._btn_simulation_path.setText(QCoreApplication.translate("_dlg_new_simulation", "Select Folder", None))
        self._lbl_simulation_name.setText(QCoreApplication.translate("_dlg_new_simulation", "Projet name :", None))
        self._lbl_simulation_path.setText(QCoreApplication.translate("_dlg_new_simulation", "Project folder :", None))
        self._lbl_scheduler.setText(QCoreApplication.translate("_dlg_new_simulation", "Scheduler :", None))
        self._lbl_simulation_timing.setText(QCoreApplication.translate("_dlg_new_simulation", "Duration time [s] :", None))
        self._lbl_simulation_step.setText(QCoreApplication.translate("_dlg_new_simulation", "Step time [s] :", None))

    @pyqtSlot()
    def _btn_accept_click (self) -> None :
        """Est la méthode appelée lorsque l'utilisateur clique sur le bouton accepter."""

        scheduler_data : SchedulerData = self._cmb_scheduler.currentData()  
        scheduler_params : SchedulerParams = SchedulerParams(self._spn_simulation_timing.value(), self._spn_simulation_step.value())
        scheduler_model : SchedulerModel = SchedulerModel(scheduler_data.scheduler_library, scheduler_data.scheduler_name, scheduler_params)
        self._value = SimulationModel(self._fld_simulation_name.text(), self._fld_simulation_path.text(), '', date.today(), 0.1, scheduler_model, [])
        self.close()



from typing import List

from PyQt5 import QtGui, QtCore
from PyQt5 import QtWidgets
from PyQt5.QtCore import Qt, QRectF, QSizeF, QPointF
from PyQt5.QtGui import QPainter
from PyQt5.QtWidgets import QGraphicsObject, QGraphicsScene, QWidget, QGraphicsView, QDialog, QStyleOptionGraphicsItem, QGraphicsItem

class FakeBox (QGraphicsObject) :

    def __init__ (self) :
        
        self._size : QSizeF = QSizeF(0, 0)
        QGraphicsObject.__init__(self, parent = None)

        self.setFlag(QGraphicsItem.ItemIsSelectable)
        self.setFlag(QGraphicsItem.ItemIsMovable)
        self.setFlag(QGraphicsItem.ItemSendsScenePositionChanges)
        self._rect = QRectF(QPointF(0, 0), self._size)

    def boundingRect(self) -> QRectF :
        return self._rect

    def paint(self, painter: QPainter, option: QStyleOptionGraphicsItem, widget) -> None:
        painter.drawRect(self.boundingRect())

    def size (self) -> QSizeF :
        return self._size

    def setSize (self, size_ : QSizeF) :
        self._size = size_

class FakeLink (QGraphicsObject) :

    def __init__ (self, parent : FakeBox, connected_to : FakeBox) :
         
        QGraphicsObject.__init__(self, parent)
        self.setParent(parent)

        self._connected_to = connected_to
        self._parent = parent
        
    def boundingRect(self) -> QRectF :
        
        self._p1 = self._parent.scenePos()
        self._p2 = self._connected_to.scenePos()
        self._line = self._p2 - self._p1

        print(self.print_values('parent', self._parent))
        print(self.print_values('connected', self._connected_to))
        print('starting point :', self.print_point(self._p1))
        print('stop point :', self.print_point(self._p2))
        print('==========================================================================')

        return QRectF(QPointF(0, 0), QSizeF(self._line.x(), self._line.y()))

    def paint(self, painter: QPainter, option: QStyleOptionGraphicsItem, widget) -> None:
        self.setRotation(-self._parent.rotation())
        path = QtGui.QPainterPath(QPointF(0, 0))
        path.lineTo(self._line)
        painter.drawPath(path)

    def print_values (self, name : str, box : FakeBox) -> None : 
        return name + ' : ' + self.print_point(box.pos()) + ' ' + self.print_size(box.size())

    def print_rect (self, rect : QRectF) -> str : 
        return '[' +  str(rect.x()) + '; ' + str(rect.y()) + '] [' + str(rect.width()) + '; ' + str(rect.height()) + ']' 

    def print_point (self, point : QPointF) -> str :
        return '[' +  str(point.x()) + '; ' + str(point.y()) + ']'

    def print_size (self, size : QSizeF) -> str :
        return '[' + str(size.width()) + '; ' + str(size.height()) + ']'


class Scenery (QGraphicsScene) :

    def __init__ (self, parent : QWidget = None, graphical_scenery = None) :
        QGraphicsScene.__init__(self, parent)
        self._graphical_scenery = graphical_scenery
        self._start_sliding_point = None
        self._boxes : List[FakeBox] = []

    def append_element (self, box : FakeBox) -> None :
        self.addItem(box)
        self._boxes.append(box)

class GraphicalScenery (QGraphicsView) :

    def __init__ (self, parent : QWidget) :

        QGraphicsView.__init__(self, parent)
        self.setScene(Scenery(self, self))

    def scenery (self) -> Scenery :
        return self.scene()

class UI (QDialog) :

    def __init__ (self, parent : QWidget = None) :

        QDialog.__init__(self, parent)
        self.setFixedSize(630, 480)
        self.setWindowTitle('Test graphics')

        self._red_brush : QtGui.QBrush = QtGui.QBrush(Qt.red)
        self._blue_brush : QtGui.QBrush = QtGui.QBrush(Qt.blue)
        self._black_pen : QtGui.QPen = QtGui.QPen(Qt.black)
        self._black_pen.setWidth(6)

        
        self._graphics_view : QGraphicsView = GraphicalScenery(self)
        self._graphics_view.setGeometry(QtCore.QRect(10, 10, 610, 460))

        box1 = FakeBox()
        box1.setPos(QPointF(10, 10))
        box1.setSize(QSizeF(100, 150))
        box1.setRotation(180)
        self._graphics_view.scenery().append_element(box1)

        box2 = FakeBox()
        box2.setPos(QPointF(300, 10))
        box2.setSize(QSizeF(100, 150))
        self._graphics_view.scenery().append_element(box2)
        
        output = FakeLink(box1, box2)


if __name__ == "__main__":
    import sys

    app = QtWidgets.QApplication(sys.argv)
    ui = UI()
    ui.show()
    sys.exit(app.exec_())


# General import
from typing import Iterable

# PyQt import
from PyQt5.QtCore import QPointF
from PyQt5.QtWidgets import QGraphicsItem

# Project import
from muphyn.packages.interface.shapes.basic_shape import GroupedShapes, Path

class ConnectionModel(GroupedShapes):

    def __init__(self, points: Iterable[QPointF], parent: QGraphicsItem = None) -> None:
        minX = min([point.x() for point in points])
        minY = min([point.y() for point in points])
        topLeft = QPointF(minX, minY)

        points = [point - topLeft for point in points]

        super().__init__(topLeft, parent)

        # Buidl steps
        steps = [Path.Step(point) for point in points]

        # Build path
        self.path = Path(steps, parent=self)


#-----------------------------------
# Imports
#-----------------------------------

# General Imports
from typing import List

# PyQt5 Imports
from PyQt5.QtCore import QPointF

# Project Imports
from muphyn.packages.interface.graphical_actions.abstract_diagram_action import AbstractDiagramAction
from muphyn.packages.interface.models.graphical_models.abstract_graphical_element import AbstractGraphicalElement

#-----------------------------------
# Class
#-----------------------------------

class DiagramMoveGraphicalElementAction (AbstractDiagramAction) :
    """Est l'action qui permet de déplacer des éléments graphiques."""

    # -------------
    # Constructors
    # -------------

    def __init__ (self, graphical_element : AbstractGraphicalElement, selection_elements : List[AbstractGraphicalElement], old_position : QPointF, new_position : QPointF) :
        
        AbstractDiagramAction.__init__(self, graphical_element.diagram_model)

        self._graphical_element_index = graphical_element.graphical_index
        self._selection_indices : List[int] = []

        for selected_element in selection_elements : 

            element_type = type(selected_element).__name__
            if element_type == 'BoxModel' or element_type == 'BoxInputModel' or element_type == 'BoxOutputModel' :
            
                if selected_element.graphical_index in self._selection_indices :
                    continue
            
                self._selection_indices.append(selected_element.graphical_index)


        self._old_position = old_position
        self._new_position = new_position

    # -------------
    # Properties
    # -------------

    def graphical_index (self) -> int : 
        """Permet de récuperer le graphical index de l'élément qui a créé l'action."""
        return self._graphical_element_index

    @property
    def old_position (self) -> QPointF :
        """Permet de récuperer la position avant modification."""
        return self._old_position

    @property
    def new_position (self) -> QPointF :
        """Permet de récuperer la position après modification."""
        return self._new_position

    @new_position.setter
    def new_position (self, new_position_ : QPointF) -> None :
        """Permet de modifier la position qui doit être appliqué à la modification."""

        if new_position_ is None : 
            return

        self._new_position = new_position_
        
    # -------------
    # Methods
    # -------------

    def contains_index (self, graphical_index_ : int) -> bool :
        """Permet de savoir si le grpahical index passé en paramètre est contenu dans la liste."""

        for selection_index in self._selection_indices : 
            if graphical_index_ == selection_index :
                return True

        return False

    def do (self) :
        
        diffrence = self._new_position - self._old_position
        for element_index in self._selection_indices :
            
            graphical_element : AbstractGraphicalElement = self.diagram_model.get_element_by_graphical_index(element_index)

            if graphical_element is None : 
                return

            graphical_element.action_pos_semaphore = True
            graphical_element.setPos(graphical_element.pos() + diffrence)
            graphical_element.action_pos_semaphore = False

    def undo (self) :
        
        diffrence = self._new_position - self._old_position
        for element_index in self._selection_indices :
            
            graphical_element : AbstractGraphicalElement = self.diagram_model.get_element_by_graphical_index(element_index)

            if graphical_element is None : 
                return

            graphical_element.action_pos_semaphore = True
            graphical_element.setPos(graphical_element.pos() - diffrence)
            graphical_element.action_pos_semaphore = False



#-----------------------------------
# Imports
#-----------------------------------

import enum
from muphyn.packages.core.box_library.box_library_data import AbstractBoxData
from muphyn.packages.interface.box_library.library_element import LibraryElement

class BoxLibraryElementRole(enum.IntEnum):
    BOX_DATA = 1

#-----------------------------------
# Class
#-----------------------------------

class BoxLibraryElement (LibraryElement) :

    # -------------
    # Constructors
    # -------------

    def __init__ (self, box_data : AbstractBoxData) :
        LibraryElement.__init__(self, box_data.box_name)
        
        self._box_data : AbstractBoxData = box_data

    # -------------
    # Properties
    # -------------

    @property
    def box_data (self) -> AbstractBoxData :
        """Permet de récuperer les données de la box à instancier."""
        return self._box_data



import sys
from PyQt5 import QtCore
from PyQt5 import QtWidgets
from muphyn.packages.core.plci_core_data_type import DataType
from muphyn.packages.interface.models.signals_model.input_connection_model import InputConnectionModel
from muphyn.packages.interface.models.signals_model.output_connection_model import OutputConnectionModel


if __name__ == "__main__":
    import sys

    app = QtWidgets.QApplication(sys.argv)
    
    view = QtWidgets.QGraphicsView()
    scene = QtWidgets.QGraphicsScene()
    scene.setSceneRect(0, 0, 1024, 768)
    view.setScene(scene)
    
    startPoint = OutputConnectionModel('t', DataType.FLOAT, QtCore.QPointF(100, 100))
    endPoint = InputConnectionModel('t', DataType.FLOAT, QtCore.QPointF(340, 340))
    
    scene.addItem(startPoint)
    scene.addItem(endPoint)

    view.show()

    sys.exit(app.exec_())